/*
 * alias.c Handles command aliases for irc.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: alias.c,v 1.25 1994/07/26 12:12:10 mrg Exp $";
#endif

#include "irc.h"

#include "alias.h"
#include "window.h"
#include "status.h"
#include "edit.h"
#include "history.h"
#include "vars.h"
#include "ircaux.h"
#include "server.h"
#include "screen.h"
#include "input.h"
#include "names.h"
#include "server.h"
#include "output.h"
#include "names.h"

extern	char	*dcc_raw_listen _((int));
extern	char	*dcc_raw_connect _((char *, int));

extern	void	strmcat _((char *, char *, int));
extern	void	strmcat_ue _((char *, char *, int));

extern	char	*FromUserHost;

extern	int	parse_number _((char **));
static char	*evaluate_operator _((char *, char *, char *, char *, int *));
static	char	*next_unit _((char *, char *, int *, int));
static	long	randm _((long));

static	char	*alias_detected _((void));
static	char	*alias_sent_nick _((void));
static	char	*alias_recv_nick _((void));
static	char	*alias_msg_body _((void));
static	char	*alias_joined_nick _((void));
static	char	*alias_public_nick _((void));
static	char	*alias_dollar _((void));
static	char	*alias_channel _((void));
static	char	*alias_server _((void));
static	char	*alias_query_nick _((void));
static	char	*alias_target _((void));
static	char	*alias_nick _((void));
static	char	*alias_invite _((void));
static	char	*alias_cmdchar _((void));
static	char	*alias_line _((void));
static	char	*alias_away _((void));
static	char	*alias_oper _((void));
static	char	*alias_chanop _((void));
static	char	*alias_modes _((void));
static	char	*alias_buffer _((void));
static	char	*alias_time _((void));
static	char	*alias_version _((void));
static	char	*alias_currdir _((void));
static	char	*alias_current_numeric _((void));
static	char	*alias_server_version _((void));
static  char	*alias_show_userhost _((void));
static	char	*alias_show_realname _((void));
static	char	*alias_online _((void));
static	char 	*alias_idle _((void));
static	char	*alias_version_str _((void));

typedef struct
{
	char	name;
	char	*(*func) _((void));
}	BuiltIns;

static	BuiltIns built_in[] =
{
	{ '.',		alias_sent_nick },
	{ ',',		alias_recv_nick },
	{ ':',		alias_joined_nick },
	{ ';',		alias_public_nick },
	{ '$',		alias_dollar },
	{ 'A',		alias_away },
	{ 'B',		alias_msg_body },
	{ 'C',		alias_channel },
	{ 'D',		alias_detected },
	{ 'E',		alias_idle },
	{ 'F',		alias_online },
/*	{ 'G' },
*/
	{ 'H', 		alias_current_numeric },
	{ 'I',		alias_invite },
	{ 'J',		alias_version_str },
	{ 'K',		alias_cmdchar },
	{ 'L',		alias_line },
	{ 'M',		alias_modes },
	{ 'N',		alias_nick },
	{ 'O',		alias_oper },
	{ 'P',		alias_chanop },
	{ 'Q',		alias_query_nick },
	{ 'R',		alias_server_version },
	{ 'S',		alias_server },
	{ 'T',		alias_target },
	{ 'U',		alias_buffer },
	{ 'V',		alias_version },
	{ 'W',		alias_currdir },
	{ 'X',		alias_show_userhost },
	{ 'Y',		alias_show_realname },
	{ 'Z',		alias_time },
	{ (char) 0,	 NULL }
};

	char	command_line[BIG_BUFFER_SIZE+1];

extern  char    *channel_key _((char *));

/* the 33 "standard" functions */
char	*function_left _((char *));
char	*function_right _((char *));
char	*function_mid _((char *));
char	*function_rand _((char *));
char	*function_srand _((char *));
char	*function_time _((char *));
char	*function_stime _((char *));
char	*function_index _((char *));
char	*function_rindex _((char *));
char	*function_match _((char *));
char	*function_rmatch _((char *));
char	*function_userhost _((char *));
char	*function_strip _((char *));
char	*function_encode _((char *));
char	*function_decode _((char *));
char	*function_ischannel _((char *));
char	*function_ischanop _((char *));
char	*function_word _((char *));
char	*function_winnum _((char *));
char	*function_winnam _((char *));
char	*function_connect _((char *));
char	*function_listen _((char *));
char	*function_tdiff _((char *));
char	*function_toupper _((char *));
char	*function_tolower _((char *));
char	*function_channels _((char *));
char	*function_servers _((char *));
char	*function_curpos _((char *));
char	*function_ppid _((char *));
char	*function_pid _((char *));

/* the 41 "extended" functions */
char	*function_before _((char *));	/* return string before char */
char	*function_after _((char *));	/* return string after char */
char    *function_leftw _((char *));	/* return first X words */
char    *function_rightw _((char *));	/* return last X words */
char    *function_midw _((char *));	/* return X words starting at word Y */
char    *function_notw _((char *));	/* return string w/o word # X */
char    *function_restw _((char *));	/* returns words after/including word X */ 
char    *function_remw _((char *));	/* returns word "X" */
char    *function_insertw _((char *));	/* insert a word into string */
char    *function_chngw _((char *));	/* change word X to "Y" */
char    *function_common _((char *));	/* return all common words in "X" and "Y" */
char    *function_diff _((char *));	/* return all words not in both "X" and "Y" */
char    *function_pattern _((char *));	/* return all words that rmatch "X" */
char    *function_filter _((char *));	/* return all words that DONT rmatch "X" */
char    *function_beforew _((char *));	/* return all words before "X" */
char    *function_tow _((char *));	/* return all words upto/including "X" */
char    *function_afterw _((char *));	/* return all words after "X" */
char    *function_fromw _((char *));	/* return all words after/including "X" */
char	*function_shift _((char *));	/* removes/returns word 1 from "X" */
char	*function_unshift _((char *));	/* makes word "Y" first word in "X" */
char	*function_push _((char *));	/* removes/returns last word from "X" */
char	*function_pop _((char *));	/* makes word "Y" last word in "X" */
char    *function_copattern _((char *));	/* for each word in X that rmatches PATT,
				   return corresponding word in Y */
char    *function_rfilter _((char *));	/* return all patterns that match X */
char    *function_rpattern _((char *));	/* return all patterns that DONT match X */
char    *function_reverse _((char *));	/* reverse a string */
char    *function_revw _((char *));	/* reverse a string word by word */
char    *function_jot _((char *));	/* return a list of ints from X to Y */
char	*function_sar _((char *));	/* search and replace */
char	*function_center _((char *));	/* center string X in field width Y */
char	*function_split _((char *));	/* split string X along char Y */
char    *function_ascii _((char *));	/* returns numbers for characters */
char    *function_chr _((char *));	/* returns characters for numbers */
char	*function_key _((char *));	/* returns channel key */
char	*function_onchannel _((char *));
char	*function_chops _((char *));
char	*function_nochops _((char *));
char	*function_numonchannel _((char *));
char	*function_which _((char *));
char	*function_isalpha _((char *));
char	*function_isdigit _((char *));
char	*function_open _((char *));
char	*function_close _((char *));
char	*function_eof _((char *));
char	*function_read _((char *));
char	*function_write _((char *));
char	*function_iptoname _((char *));
char	*function_nametoip _((char *));
char	*function_convert _((char *));
char	*function_translate _((char *));
char	*function_server_version _((char *));
char	*function_unlink _((char *));
char	*function_rename _((char *));
char	*function_rmdir _((char *));
char	*function_truncate _((char *));

/* Karll's array suite */
extern char *function_indextoitem _((char *));
extern char *function_itemtoindex _((char *));
extern char *function_igetitem _((char *));
extern char *function_getitem _((char *));
extern char *function_setitem _((char *));
extern char *function_finditem _((char *));
extern char *function_matchitem _((char *));
extern char *function_rmatchitem _((char *));
extern char *function_getmatches _((char *));
extern char *function_getrmatches _((char *));
extern char *function_delitem _((char *));
extern char *function_numitems _((char *));
extern char *function_getarrays _((char *));
extern char *function_numarrays _((char *));
extern char *function_delarray _((char *));
extern char *function_ifinditem _((char *));
extern char *function_ifindfirst _((char *));

typedef struct
{
	char	*name;
	char	*(*func) _((char *));
}	BuiltInFunctions;


  static BuiltInFunctions	built_in_functions[] =
  {
        { "AFTER",              function_after },
        { "AFTERW",             function_afterw },
        { "ASCII",              function_ascii },
        { "BEFORE",             function_before },
        { "BEFOREW",            function_beforew },
 	{ "CENTER",		function_center },
        { "CHR",                function_chr },
	{ "CLOSE",		function_close },
 	{ "CONNECT",		function_connect },
	{ "CONVERT",		function_convert },
 	{ "CURPOS",		function_curpos },
 	{ "CHANUSERS",		function_onchannel },
        { "CHNGW",              function_chngw },
        { "CHOPS",              function_chops },
        { "COMMON",             function_common },
        { "COPATTERN",          function_copattern },
	{ "DECODE",		function_decode },
        { "DELARRAY",           function_delarray },
        { "DELITEM",            function_delitem },
        { "DIFF",               function_diff },
	{ "ENCODE",		function_encode },
	{ "EOF",		function_eof },
        { "FILTER",             function_filter },
        { "FINDITEM",           function_finditem },
        { "FROMW",              function_fromw },
        { "GETARRAYS",          function_getarrays },
        { "GETITEM",            function_getitem },
        { "GETMATCHES",         function_getmatches },
        { "GETRMATCHES",        function_getrmatches },
	{ "INDEX",		function_index },
        { "IFINDFIRST",         function_ifindfirst },
        { "IFINDITEM",          function_ifinditem },
        { "IGETITEM",           function_igetitem },
        { "INDEXTOITEM",        function_indextoitem },
        { "INSERTW",            function_insertw },
	{ "IPTONAME",		function_iptoname },
	{ "ISALPHA",		function_isalpha },
	{ "ISCHANNEL",		function_ischannel },
	{ "ISCHANOP",		function_ischanop },
	{ "ISDIGIT",		function_isdigit },
        { "ITEMTOINDEX",        function_itemtoindex },
        { "JOT",                function_jot },
        { "KEY",                function_key },
	{ "LEFT",		function_left },
        { "LEFTW",              function_leftw },
	{ "LISTEN",		function_listen },
	{ "MATCH",		function_match },
        { "MATCHITEM",          function_matchitem },
	{ "MID",		function_mid },
        { "MIDW",               function_midw },
	{ "MYCHANNELS",		function_channels },
	{ "MYSERVERS",		function_servers },
	{ "NAMETOIP",		function_nametoip },
        { "NOCHOPS",            function_nochops },
        { "NOTW",               function_notw },
        { "NUMARRAYS",          function_numarrays },
        { "NUMITEMS",           function_numitems },
	{ "NUMONCHANNEL",	function_numonchannel },
        { "ONCHANNEL",          function_onchannel },
	{ "OPEN",		function_open },
        { "PATTERN",            function_pattern },
	{ "PID",		function_pid },
	{ "POP",		function_pop },
	{ "PPID",		function_ppid },
	{ "PUSH",		function_push },
	{ "RENAME",		function_rename },
  	{ "RIGHT",		function_right },
  	{ "RAND",		function_rand },
	{ "READ",		function_read },
        { "REVERSE",            function_reverse },
        { "REVW",               function_revw },
        { "RFILTER",            function_rfilter },
        { "RIGHTW",             function_rightw },
	{ "RINDEX",		function_rindex },
        { "REMW",               function_remw },
        { "RESTW",              function_restw },
	{ "RMATCH",		function_rmatch },
        { "RMATCHITEM",         function_rmatchitem },
	{ "RMDIR",		function_rmdir },
        { "RPATTERN",           function_rpattern },
	{ "SAR",		function_sar },
        { "SETITEM",            function_setitem },
	{ "SHIFT",		function_shift },
	{ "SPLIT",		function_split },
  	{ "SRAND",		function_srand },
	{ "STIME",		function_stime },
	{ "STRIP",		function_strip },
	{ "TDIFF",		function_tdiff },
  	{ "TIME",		function_time },
	{ "TOLOWER",		function_tolower },
	{ "TOUPPER",		function_toupper },
        { "TOW",                function_tow },
	{ "TR",			function_translate },
	{ "TRUNC",		function_truncate },
	{ "UNLINK",		function_unlink },
	{ "UNSHIFT",		function_unshift },
  	{ "USERHOST",		function_userhost },
	{ "VERSION",		function_server_version },
	{ "WHICH",		function_which },
  	{ "WINNUM",		function_winnum },
  	{ "WINNAM",		function_winnam },
	{ "WORD",		function_word },
	{ "WRITE",		function_write },
  	{ (char *) 0,		NULL }
  };
  
/* alias_illegals: characters that are illegal in alias names */
	char	alias_illegals[] = " #+-*/\\()={}[]<>!@$%^~`,?;:|'\"";

Alias	*alias_list[] =
{
	(Alias *) 0,
	(Alias *) 0
};

/* alias_string: the thing that gets replaced by the $"..." construct */
static	char	*alias_string = (char *) 0;

static	int	eval_args;

/* function_stack and function_stkptr - hold the return values from functions */
static	char	*function_stack[128] =
{ 
	(char *) 0
};
static	int	function_stkptr = 0;

extern	char	*MatchingBracket _((char *, char, char));
extern	char	*MatchingBracketBackwards _((char *, char *, char, char));




/* makes foo[one][two] look like tmp.one.two -- got it? */
#ifdef __STDC__
void remove_brackets (char *name, char *args, int *arg_flag)
#else
void remove_brackets (name, args, arg_flag)
char *name;
char *args;
int *arg_flag;
#endif
{
	char *str = NULL, *ptr, *right;
	char *result1;

	while (ptr = strchr(name, '['))
	{
		*ptr++ = 0;				/* name <- ptr */
		malloc_strcpy(&str, ptr);		/* str */
		right = str + 1;			/* str <- right */
		ptr = MatchingBracket(right, '[', ']'); /* str <- ptr */
		if (ptr)				/* str <- ptr */
			*ptr++ = 0;			/* str <- ptr */

		result1 = expand_alias(NULL, right, args, arg_flag, NULL);
		strcat(name, ".");
		strcat(name, result1);
		strcat(name, ptr);
		new_free(&result1);
		new_free(&str);
	}
}

/*
 * find_alias: looks up name in in alias list.  Returns the Alias entry if
 * found, or null if not found.   If unlink is set, the found entry is
 * removed from the list as well.  If match is null, only perfect matches
 * will return anything.  Otherwise, the number of matches will be returned. 
 */
#ifdef __STDC__
extern	Alias	*find_alias (Alias **list, char *name, int unlink, int *match)
#else
extern	Alias	*find_alias (list, name, unlink, match)
	Alias	**list;
	char	*name;
	int	unlink;
	int	*match;
#endif
{
	Alias	*tmp,
		*last = (Alias *) 0;
	int	cmp,
		len;
	int (*cmp_func) ();
	char	*bptr;

	if (match)
	{
		*match = 0;
		cmp_func = my_strnicmp;
	}
	else
		cmp_func = my_stricmp;

	if (name)
	{
		/* The function that is looking for the variable should not
		   be responsible for resolving any array variables.  That
		   should be and is our job.  So anyone can pass us a
		   variable name like 'foo[$one][two]' and we know just
		   what to do with it.  righto?  */
		int af;
		remove_brackets(name, NULL, &af);

		len = strlen(name);
		for (tmp = *list; tmp; tmp = tmp->next)
		{
			if ((cmp = cmp_func(name, tmp->name, len)) == 0)
			{
				static already_looking = 0;
				if (tmp->stub)
				{
					/* If already_looking is 1, then
					 * we are un-stubbing this alias
					 * because we are loading a file
					 * that presumably it must be in.
					 * So we dont load it again (duh).
					 */
					if (!already_looking)
					{
						already_looking = 1;
						load("LOAD", tmp->stub, empty_string);
						already_looking = 0;
					}
					/* it is very important for every
					 * caller of this function to assume
					 * that 'stub' is always NULL unless
					 * they are checking for non-NULLness.
					 * NO NEED to ever free or reassign
					 * the stub field anywhere but here!
					 */
					/* Whee.. recursion implies that
					 * the interesting parts of tmp have
					 * already been re-assigned by the 
					 * load, so we just clean up after
					 * ourselves.  We have to complete
					 * this call, becuase whoever called
					 * us is expecting a reply, so the load
					 * was just a temporary time-out while
					 * we feteched the interesting alias.
					 */
					new_free(&(tmp->stub));
					tmp->stub = NULL;

					/* I dont know if this will get
					   me in trouble or not, but im 
					   going to try it.  If we have
					   loaded the file, and this alias
					   is therefore de-stubbed, if
					   it wasnt assigned somewhere in
					   the file, then we should unlink
					   the stub. */
					if (!tmp->stuff)
						unlink = 1;
				}

				if (unlink)
				{
					if (last)
						last->next = tmp->next;
					else
						*list = tmp->next;
				}
				if (match)
				{
					(*match)++;
					if (strlen(tmp->name) == len)
					{
						*match = 0;
						return (tmp);
					}
				}
				else
					return (tmp);
			}
			else if (cmp < 0)
				break;
			last = tmp;
		}
	}
	if (match && (*match == 1))
		return (last);
	else
		return ((Alias *) 0);
}

/*
 * insert_alias: adds the given alias to the alias list.  The alias list is
 * alphabetized by name 
 */
#ifdef __STDC__
extern	void	insert_alias (Alias **list, Alias *alias)
#else
extern	void	insert_alias (list, alias)
	Alias	**list;
	Alias	*alias;
#endif
{
	Alias	*tmp,
		*last,
		*foo;

	last = (Alias *) 0;
	for (tmp = *list; tmp; tmp = tmp->next)
	{
		if (strcmp(alias->name, tmp->name) < 0)
			break;
		last = tmp;
	}
	if (last)
	{
		foo = last->next;
		last->next = alias;
		alias->next = foo;
	}
	else
	{
		alias->next = *list;
		*list = alias;
	}
}

/*
 * add_alias: given the alias name and "stuff" that makes up the alias,
 * add_alias() first checks to see if the alias name is already in use... if
 * so, the old alias is replaced with the new one.  If the alias is not
 * already in use, it is added. 
 */
#ifdef __STDC__
void	add_alias (int type, char *name, char *stuff)
#else
void	add_alias (type, name, stuff)
	int	type;
	char	*name,
		*stuff;
#endif
{
	Alias	*tmp;
	char	*ptr;

	upper(name);
	if (type == COMMAND_ALIAS)
		say("Alias	%s added", name);
	else
	{
		if (!strcmp(name, "FUNCTION_RETURN"))
		{
			if (function_stack[function_stkptr])
				new_free(&function_stack[function_stkptr]);
			malloc_strcpy(&function_stack[function_stkptr], stuff);
			return;
		}
		if ((ptr = sindex(name, alias_illegals)) != NULL)
		{
			yell("Assign names may not contain '%c'", *ptr);
			return;
		}
		say("Assign %s added", name);
	}

	/* dont forget -- if 'name' contains any [] pairs, find_alias will
	   evaluate and remove them for us... */
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) 0)) == (Alias *) 0)
	{
		tmp = (Alias *) new_malloc(sizeof(Alias));
		if (tmp == (Alias *) 0)
		{
			yell("Couldn't allocate memory for new alias!");
			return;
		}
		tmp->name = (char *) 0;
		tmp->stuff = (char *) 0;
		tmp->stub = (char *) 0;
	}
	malloc_strcpy(&(tmp->name), name);
	malloc_strcpy(&(tmp->stuff), stuff);
	tmp->mark = 0;
	tmp->global = loading_global;
	insert_alias(&(alias_list[type]), tmp);
}

/*
 * add_stub: You give it an alias name and a filename, and it generates
 * a stub for that alias.  If an alias is already in place for that name,
 * it is duly clobbered.  If it isnt already in place, this makes a new one.
 */
#ifdef __STDC__
void	add_stub (int type, char *name, char *file)
#else
void	add_stub (type, name, file)
	int	type;
	char	*name,
		*file;
#endif
{
	Alias	*tmp;
	char	*ptr;

	upper(name);
	if (type == COMMAND_ALIAS)
		say("Alias	%s stubbed to file %s", name, file);
	else
	{
		if (!strcmp(name, "FUNCTION_RETURN"))
		{
			yell("You cannot stub the FUNCTION_RETURN variable");
			return;
		}
		if ((ptr = sindex(name, alias_illegals)) != NULL)
		{
			yell("Assign names may not contain '%c'", *ptr);
			return;
		}
		say("Assign	%s stubbed to file %s", name, file);
	}
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) 0)) ==
			(Alias *) 0)
	{
		tmp = (Alias *)new_malloc(sizeof(Alias));
		if (tmp == (Alias *) 0)
		{
			yell("Couldn't allocate memory for new alias!");
			return;
		}
		tmp->name = (char *) 0;
		tmp->stuff = (char *) 0;
		tmp->stub = (char *) 0;
	}
	malloc_strcpy(&(tmp->name), name);
	malloc_strcpy(&(tmp->stub), file);
	new_free(&(tmp->stuff));
	tmp->stuff = NULL;
	tmp->mark = 0;
	tmp->global = loading_global;
	insert_alias(&(alias_list[type]), tmp);
}



/* alias_arg: a special version of next_arg for aliases */
#ifdef __STDC__
static	char	*alias_arg (char **str, u_int *pos)
#else
static	char	*alias_arg (str, pos)
	char	**str;
	u_int	*pos;
#endif
{
	char	*ptr;

	if (!*str)
		return (char *) 0;
	*pos = 0;
	ptr = *str;
	while (' ' == *ptr)
	{
		ptr++;
		(*pos)++;
	}
	if (*ptr == '\0')
	{
		*str = empty_string;
		return ((char *) 0);
	}
	if ((*str = sindex(ptr, " ")) != NULL)
		*((*str)++) = '\0';
	else
		*str = empty_string;
	return (ptr);
}

/* canon_number: canonicalizes number to something relevant */
/* If FLOATING_POINT_MATH isnt set, it truncates it to an integer */
#ifdef __STDC__
extern void canon_number (char *input)
#else
extern void canon_number (input)
char *input;
#endif
{
	int end = strlen(input);

	if (end)
		end--;
	else
		return;		/* nothing to do */

	if (get_int_var(FLOATING_POINT_MATH_VAR))
	{
		/* remove any trailing zeros */
		while (input[end] == '0')
			end--;

		/* If we removed all the zeros and all that is
		   left is the decimal point, remove it too */
		if (input[end] == '.')
			end--;

		input[end+1] = 0;
	}
	else
	{
		char *dot = index(input, '.');
		if (dot)
			*dot = 0;
	}
}

/* word_count: returns the number of words in the given string */
#ifdef __STDC__
extern	int	word_count (char *str)
#else
extern	int	word_count (str)
	char	*str;
#endif
{
	int	cnt = 0;
	char	*ptr;

	while (1)
	{
		if ((ptr = sindex(str, "^ ")) != NULL)
		{
			cnt++;
			if ((str = sindex(ptr, " ")) == (char *) 0)
				return (cnt);
		}
		else
			return (cnt);
	}
}

#ifdef __STDC__
static	char	*built_in_alias (char c)
#else
static	char	*built_in_alias (c)
	char	c;
#endif
{
	BuiltIns	*tmp;
	char	*ret = (char *) 0;

	for(tmp = built_in;tmp->name;tmp++)
		if (c == tmp->name)
		{
			malloc_strcpy(&ret, tmp->func());
			break;
		}
	return(ret);
}

/*
 * find_inline: This simply looks up the given str.  It first checks to see
 * if its a user variable and returns it if so.  If not, it checks to see if
 * it's an IRC variable and returns it if so.  If not, it checks to see if
 * its and environment variable and returns it if so.  If not, it returns
 * null.  It mallocs the returned string 
 */
#ifdef __STDC__
static	char	*find_inline (char *str)
#else
static	char	*find_inline (str)
	char	*str;
#endif
{
	Alias	*alias;
	char	*ret = NULL;
	char	*tmp;

	if ((alias = find_alias(&(alias_list[VAR_ALIAS]), str, 0, (int *) NULL))
			!= NULL)
	{
		malloc_strcpy(&ret, alias->stuff);
		return (ret);
	}
	if ((strlen(str) == 1) && (ret = built_in_alias(*str)))
		return(ret);
	if ((ret = make_string_var(str)) != NULL)
		return (ret);
#ifdef DAEMON_UID
	if (getuid() == DAEMON_UID)
	malloc_strcpy(&ret, (getuid() != DAEMON_UID) && (tmp = getenv(str)) ?
				tmp : empty_string);
#else
	malloc_strcpy(&ret, (tmp = getenv(str)) ? tmp : empty_string);
#endif /* DAEMON_UID */
	return (ret);
}

#ifdef __STDC__
static	char	*call_function (char *name, char *f_args, char *args, int *args_flag)
#else
static	char	*call_function (name, f_args, args, args_flag)
	char	*name,
		*f_args,
		*args;
	int	*args_flag;
#endif
{
	char	*tmp;
	char	*result = (char *) 0;
	char	*sub_buffer = (char *) 0;
	int	builtnum;
	char	*debug_copy = (char *) 0;

	tmp = expand_alias((char *) 0, f_args, args, args_flag, NULL);
	if (get_int_var(DEBUG_VAR) & DEBUG_FUNCTIONS)
		malloc_strcpy(&debug_copy, tmp);
	for (builtnum = 0; built_in_functions[builtnum].name != NULL &&
			my_stricmp(built_in_functions[builtnum].name, name);
			builtnum++)
		;
	if (built_in_functions[builtnum].name)
		result = built_in_functions[builtnum].func(tmp);
	else
	{
		sub_buffer = new_malloc(strlen(name)+strlen(tmp)+2);
		strcpy(sub_buffer, name);
		strcat(sub_buffer, " ");
		strcat(sub_buffer, tmp);
		function_stack[++function_stkptr] = (char *) 0;
		parse_command(sub_buffer, 0, empty_string);
		new_free(&sub_buffer);
		eval_args=1;
		result = function_stack[function_stkptr];
		function_stack[function_stkptr] = (char *) 0;
		if (!result)
			malloc_strcpy(&result, empty_string);
		function_stkptr--;
	}
	if (debug_copy)
	{
		yell("Function %s(%s) returned %s",
		    name, debug_copy, result);
		new_free(&debug_copy);
	}
	new_free(&tmp);
	return result;
}


/* Given a pointer to an operator, find the last operator in the string */
#ifdef __STDC__
char	*lastop (char *ptr)
#else
char	*lastop (ptr)
	char	*ptr;
#endif
{
	while (ptr[1] && index("!=<>&^|#+/%,-*", ptr[1]))
		ptr++;
	return ptr;
}

#ifdef __STDC__
int isoperator(char ptr)
#else
int isoperator(ptr)
char ptr;
#endif
{
	return (index("!=<>&^|#+/%,-*", ptr)) ? 1 : 0;
}

/* Not yet */
#ifdef NEW_EVAL
/* infix_to_postfix -- translate an infix expression to a postfix
 * expression. ick.
 */
#ifdef __STDC__
char	*infix_to_postfix(char *ptr)
#else
char	*infix_to_postfix(ptr)
char *ptr;
#endif
{
	char *answer = NULL;
	char *convertstack[MAX_STACKSIZE];
	int stackptr = 0;

	/* This is a two pass compilation.  First pass we seek and 
	 * destroy any unary ops
	 */
	/* Remove the following unary forms:
		++foo 	foo++		--foo 	foo--
		+foo			-foo
	 */
	while (ptr && *ptr)
	{
		char *nextop = find_next_operator(ptr);

		/* we have to rid the string of unary operators. ick. */
		if (nextop == ptr)
		{
			if ((ptr[0] == '+' && ptr[1] == '+') ||
			    (ptr[0] == '-' && ptr[1] == '-'))
			{
				char save;
				nextop = find_next_operator(ptr+2);
				save = *nextop;
				*nextop = 0;
				variable = find_inline(ptr+2);
				*nextop = save;
				value = atol(variable);
				if (ptr[0] == '+')
					value++;
				else
					value--;
				sprintf(value2, "%ld", value);
				add_alias(ASSIGN_ALIAS, ptr+2, value2);
				strcpy(ptr, ptr+2);
			}

			else if (ptr[0] == '+')
				strcpy(ptr, ptr+1);

			else if (ptr[0] == '-')
			{
				pushstack(&convertstack, &stackptr, "0");
				pushstack(&convertstack, &stackptr, ptr+1);
				pushstack(&convertstack, &stackptr, "-");
				strcpy(ptr, ptr+1);
			}
			continue;
		}
		else	/* check to see if its foo-- or foo++ */
		{
			if (  (nextop[0] == '-' || nextop[0] == '+')
			   && (nextop[0] == nextop[1])
			   && (isspace(nextop[2]) || isoperator(nextop[2])))
			{
				char save;
				save = *nextop;
				*nextop = 0;
				variable = find_inline(ptr);
				*nextop = save;
				value = atol(variable);
				if (ptr[0] == '+')
					value++;
				else
					value--;
				sprintf(value2, "%ld", value);
				add_alias(ASSIGN_ALIAS, ptr+2, value2);
				nextop[0] = (nextop[0] == '-') ? '+' : '-';
				nextop[1] = '1';
			}
	}
	/* strip out all the spaces */
	for (foo = ptr; foo && *foo; foo++)
	{
		while (isspace(*foo))
			strcpy(foo, foo + 1);
	}

	/* now we parse the expression */
	{
	int stage;
	for (stage = NU_UNIT; stage >= NU_EXPR; stage--)
	{

}

/* evaluate_infix evaluates an infix notation (duh. ;-) */
#ifdef __STDC__
char 	*evaluate_infix(char *ptr, char *args, int *arg_flag)
#else
char 	*evaluate_infix (ptr, args, arg_flag)
char *ptr, args;
int *arg_flag;
#endif
{
	char *pfix = infix_to_postfix (ptr);
	char *answer = evaluate_postfix (pfix, args, arg_flag);
	new_free(&pfix);
	return answer;
}

/* Expand does whats neccesary to expand some expression text
   If something is within []s, it is extracted and expanded.
   Otherwise, the text is assumed to be a variable and is expanded
 */
#ifdef __STDC__
char *expand(char *ptr, char *args, int *arg_flag)
#else
char *expand(ptr, args, arg_flag)
char *ptr;
char *args;
int *arg_flag;
#endif
{
	if (*ptr == '[')
	{
		char *str = MatchingBracket(ptr+1, '[', ']');
		*str = 0;
		result1 = expand_alias(NULL, ptr+1, args, arg_flag, NULL);
	}
	else
		result1 = find_inline(ptr);

	return result1;
}

/*
	Evaluate_postfix -- evaluate postfix expressions (doh!)

	The input has to be a well behaved syntax, prefereably taken
	from the "infix_to_postfix" function.  All others call at their
	own peril. 
*/
#ifdef __STDC__
char	*evaluate_postfix (char *ptr, char *args, int *arg_flag)
#else
char *evaluate_postfix (ptr, args, arg_flag)
char *ptr;
char *args;
int *arg_flag;
#endif
{
	char *stack[MAX_STACKSIZE] = {0};
	char *opone, optwo;
	int stackptr = 0;
	double v1, v2, fv;
	int i1, i2, fi;
	char *value;

	while (ptr && *ptr)
	{
		str = next_arg(ptr, &ptr);

		if (isoperator(*str))
		{
			if (popstack(&stack, &stackptr, &opone) == 0)
			{
				yell("Expression syntax error");
				fv = 0;
			}
			else if (popstack(&stack, &stackptr, &optwo) == 0)
			{
				yell("Expression syntax error");
				fv = 0;
			}
			else
			{

switch (*str)
{
	case '+':
	case '-':
	{
		ropone = expand(opone);
		roptwo = expand(optwo);
		v1 = atof(ropone);
		v2 = atof(roptwo);
		if (*str == '+')
			fv = v2 + v1;
		else
			fv = v2 - v1;
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
	case '#':
	{
		ropone = expand(opone);
		roptwo = expand(optwo);
		if (str[1] == '#')
		{
			malloc_strcpy(&value, ropone);
			malloc_strcat(&value, roptwo);
		}
		break;
	}
	case '*':
	case '%':
	case '/':
	{
		ropone = expand(opone);
		roptwo = expand(optwo);
		v1 = atof(opone);
		v2 = atof(optwo);

		if ((*str != '*') && (v1 == 0.0))
		{
			yell("Invalid syntax: divide or mod by zero");
			fv = 0;
		}
		else if (*str == '/')
			fv = v2 / v1;
		else if (*str == '%')
			fv = (float)((int)v2 % (int)v1);
		else
			fv = v2 * v1;
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
	case '&':
	{
		if (str[1] == '&')
		{
			roptwo = expand(optwo);
			if (*roptwo && (*roptwo != '0'))
			{
				ropone = expand(opone);
				i1 = atol(ropone);
				i2 = atol(roptwo);
				fv = (float)(i2 && i1);
				malloc_sprintf(&value, "%lf", fv);
				canon_number(value);
				break;
			}
			else
				fv = 0.0;
		}
		else
		{
			ropone = expand(opone);
			roptwo = expand(optwo);
			i1 = atol(ropone);
			i2 = atol(roptwo);
			fv = (float)(i2 & i1);
		}
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
	case '|':
	{
		if (str[1] == '|')
		{
			roptwo = expand(optwo);
			if (!*roptwo && (*roptwo == '0'))
			{
				ropone = expand(opone);
				i1 = atol(ropone);
				i2 = atol(roptwo);
				fv = (float)(i2 || i1);
				malloc_sprintf(&value, "%lf", fv);
				canon_number(value);
				break;
			}
			else
				fv = 1.0;
		}
		else
		{
			ropone = expand(opone);
			roptwo = expand(optwo);
			v1 = atol(ropone);
			v2 = atol(roptwo);
			fv = (float)(i2 | i1);
		}
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
	case '^':
	{
		ropone = expand(opone);
		roptwo = expand(optwo);
		if (str[1] == '^')
		{
			i1 = atol(ropone) ? 1 : 0;
			i2 = atol(roptwo) ? 1 : 0;
			fv = (float)(i2 ^ i1);
		}
		else
		{
			i1 = atol(ropone);
			i2 = atol(roptwo);
			fv = (float)(i2 ^ i1);
		}
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
	case '=':
	{
		if (ptr[1] == '=')
		{
			ropone = expand(opone);
			roptwo = expand(optwo);
			fv = my_stricmp(ropone, roptwo) ? 0 : 1;
		}
		else
		{
			ropone = expand_alias((char *) 0, optwo, args, arg_flag, NULL);
			roptwo = expand(opone);
			while (*ropone && *ropone == ' ' && ropone < ptr)
				ropone++;
			if (ropone < ptr)
				add_alias(VAR_ALIAS, ropone, roptwo);
			else
				yell("Illegal assignment");
		}
	}
	case '!':
	{
		ropone = expand(optwo);
		if (ropone && (*ropone != '0'))
			malloc_strcpy(&value, "0");
		else
			malloc_strcpy(&value, "1");
		break;
	}
	case '~':
	{
		ropone = expand(optwo);
		i1 = atol(ropone);
		i1 = ~i1;
		malloc_sprintf("%ld", i1);
		break;
	}
	case ',':
	{
		roptwo = expand(optwo);
		ropone = expand(opone);
		malloc_strcpy(&value, ropone);
		break;
	}
	case '?':
	{
		char *partone = opone;
		char *partwo  = index(opone, ':');
		if (partwo == (char *)0)
		{
			yell("Expression syntax error: ? with no :");
			fv = 0;
			malloc_sprintf(&value, "%lf", fv);
			canon_number(value);
			break;
		}

		roptwo = expand(optwo);
		if (*roptwo && (*roptwo != '0'))
		{
			ropone = expand(partone);
			v1 = atof(partone);
		}
		else
		{
			ropone = expand(partwo);
			v1 = atof(partwo);
		}
		fv = v1;
		malloc_sprintf(&value, "%lf", fv);
		canon_number(value);
		break;
	}
}
				/* leave here for after switch statement */
				new_free(&ropone);
				new_free(&roptwo);
/* How will we handle static? */
				pushstack(&stack, &stackptr, &value);
			}
		}
		else
		{
			pushstack(&stack, &stackptr, str);
		}
	}
	if (stackptr)
	{
		yell("Expression syntax error");
		sprintf(value, "0", fv);
	}

	return m_strdup(value);
}
#else /* not yet */

#define	NU_EXPR	0
#define	NU_ASSN	1
#define NU_TERT 2
#define NU_CONJ 3
#define NU_BITW 4 
#define NU_COMP 5 
#define NU_ADD  6 
#define NU_MULT 7 
#define NU_UNIT 8

#ifdef __STDC__
static	char	*next_unit (char *str, char *args, int *arg_flag, int stage)
#else
static	char	*next_unit (str, args, arg_flag, stage)
	char	*str,
		*args;
	int	*arg_flag,
		stage;
#endif
{
	char	*ptr,
		*ptr2,
		*right;
	int	got_sloshed = 0;
	char	*lastc;
	char	tmp[40];
	char	*result1 = (char *) 0,
		*result2 = (char *) 0;
	long	value1 = 0,
		value2,
		value3 = 0;
	double	dvalue1 = 0.0,
		dvalue2 = 0.0,
		dvalue3 = 0.0;
	char	op;
	int	display;
	char	*ArrayIndex,
		*EndIndex;
	char	*firstchar = str;

	while (isspace(*str))
		++str;
	if (!*str)
	{
		malloc_strcpy(&result1, empty_string);
		return result1;
	}
	lastc = str+strlen(str)-1;
	while (isspace(*lastc))
		*lastc-- = '\0';

	if (stage == NU_UNIT && *lastc == ')' && *str == '(')
	{
		str++, *lastc-- = '\0';
		return next_unit(str, args, arg_flag, NU_EXPR);
	}
	if (!*str)
	{
		malloc_strcpy(&result1, empty_string);
		return result1;
	}
	for (ptr = str; *ptr; ptr++)
	{
		if (got_sloshed)
		{
			got_sloshed = 0;
			continue;
		}

		switch(*ptr)
		{
		case '\\':
			got_sloshed = 1;
			continue;
		case '(':
			if (stage != NU_UNIT || ptr == str)
			{
				if (!(ptr2 = MatchingBracket(ptr+1, '(', ')')))
					ptr = ptr+strlen(ptr)-1;
				else
					ptr = ptr2;
				break;
			}
			*ptr++ = '\0';
			right = ptr;
			ptr = MatchingBracket(right, LEFT_PAREN, RIGHT_PAREN);
			if (ptr)
				*ptr++ = '\0';
			result1 = call_function(str, right, args, arg_flag);
			if (ptr && *ptr)
			{
				malloc_strcat(&result1, ptr);
				result2 = next_unit(result1, args, arg_flag, stage);
				new_free(&result1);
				result1 = result2;
			}
			return result1;
		case '[':
			if (stage != NU_UNIT)
			{
				if (!(ptr2 = MatchingBracket(ptr+1, '[', ']')))
					ptr = ptr+strlen(ptr)-1;
				else
					ptr = ptr2;
				break;
			}
			*ptr++ = '\0';
			right = ptr;
			ptr = MatchingBracket(right, LEFT_BRACKET, RIGHT_BRACKET);
			if (ptr)
				*ptr++ = '\0';
			result1 = expand_alias((char *) 0, right, args, arg_flag, NULL);
			if (*str)
			{
				result2 = new_malloc(strlen(str)+
						(result1?strlen(result1):0)+
						(ptr?strlen(ptr):0) + 2);
				strcpy(result2, str);
				strcat(result2, ".");
				strcat(result2, result1);
				new_free(&result1);
				if (ptr && *ptr)
				{
					strcat(result2, ptr);
					result1 = next_unit(result2, args, arg_flag, stage);
				}
				else
				{
					result1 = find_inline(result2);
					if (!result1)
						malloc_strcpy(&result1, empty_string);
				}
				new_free(&result2);
			}
			else if (ptr && *ptr)
			{
				malloc_strcat(&result1, ptr);
				result2 = next_unit(result1, args, arg_flag, stage);
				new_free(&result1);
				result1 = result2;
			}
			/* WARNING WARNING WARNING KLUDGE HERE */
			if (!*result1)
				malloc_strcpy(&result1, empty_string);
			return result1;
		case '-':
		case '+':
                        if (*(ptr+1) == *(ptr))  /* index operator */
                        {
                                char *tptr;

                                *ptr++ = '\0';
				if (ptr == str + 1)	/* Its a prefix */
                                        tptr = str + 2;
				else			/* Its a postfix */
                                        tptr = str;

			    	result1 = find_inline(tptr);
				if (!result1) 
					malloc_strcpy(&result1,"0");

				{           /* This isnt supposed to be
                                                attached to the if, so
                                                dont "fix" it. */
                                        long r;
                                        r = atol(result1);
                                        if (*ptr == '+')
                                                r++;
                                        else    r--;
                                        sprintf(tmp,"%ld",r);
                                        display = window_display;
                                        window_display = 0;
                                        add_alias(VAR_ALIAS,tptr,tmp);
                                        window_display = display;
                                }
		/* A kludge?  Cheating?  Maybe.... */
				if (ptr == str + 1) 
				{
	                                *(ptr-1) = ' ';
					*ptr = ' ';
				} else            
				{
                                        if (*ptr == '+')
					        *(ptr-1) = '-';
                                        else
                                                *(ptr-1) = '+';
					*ptr = '1';
				}
                                ptr = str; 
                                new_free(&result1);
                                break;
                        }
			if (ptr == str) /* It's unary..... do nothing */
				break;
			if (ptr[1] == '=') /* += operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					result2 = next_unit(ptr, args, arg_flag, stage);
					window_display = 0;
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}
					result1 = find_inline(varname);
					if (!result1 || !*result1)
						dvalue1 = 0;
					else
						dvalue1 = atof(result1);	
					dvalue2 = atof(result2);
					new_free(&result1);
					new_free(&result2);
					if (op == '-')
						dvalue3 = dvalue1 - dvalue2;
					else
						dvalue3 = dvalue1 + dvalue2;
					sprintf(tmp, "%lf", dvalue3);
					canon_number(tmp);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, tmp);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result1);
					malloc_strcpy(&result2, tmp);
					return result2;
				}
			}
				
			if (stage != NU_ADD)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			dvalue1 = atof(result1);
			dvalue2 = atof(result2);
			new_free(&result1);
			new_free(&result2);
			if (op == '-')
				dvalue3 = dvalue1 - dvalue2;
			else
				dvalue3 = dvalue1 + dvalue2;
			sprintf(tmp, "%lf", dvalue3);
			canon_number(tmp);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '/':
		case '*':
  		case '%':
			if (ptr == str) /* It's unary..... do nothing */
				break;

			if (ptr[1] == '=') /* += operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					result2 = next_unit(ptr, args, arg_flag, stage);
					window_display = 0;
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}
					result1 = find_inline(varname);
					if (!result1 || !*result1)
						dvalue1 = 0.0;
					else
						dvalue1 = atof(result1);
					dvalue2 = atof(result2);
					new_free(&result1);
					new_free(&result2);
					if (op == '*')
						dvalue3 = dvalue1 * dvalue2;
					else if (op == '/')
					{
						if (dvalue2 == 0.0)
						{
							value3 = 0.0;
							yell("Division by zero!");
						}
						else
							dvalue3 = dvalue1 / dvalue2;
					}
					else if (op == '%')
					{
						if ((int)dvalue2 == 0)
						{
							dvalue3 = 0.0;
							yell("Mod by zero!");
						}
						else
							dvalue3 = (int)dvalue1 % (int)dvalue2;
					}
					sprintf(tmp, "%ld", dvalue3);
					canon_number(tmp);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, tmp);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result1);
					malloc_strcpy(&result2, tmp);
					return result2;
				}
			}
				
			if (stage != NU_MULT)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			dvalue1 = atof(result1);
			dvalue2 = atof(result2);
			new_free(&result1);
			new_free(&result2);
			if (op == '/')
			{
				if (dvalue2)
					dvalue3 = dvalue1 / dvalue2;
				else
				{
					dvalue3 = 0.0;
					say("Division by zero");
				}
			}
			else if (op == '*')
				dvalue3 = dvalue1 * dvalue2;
                        else
                        {
                                if ((int)dvalue2)
					dvalue3 = (int)dvalue1 % (int)dvalue2;
                                else
                                {
                                        value3 = 0.0;
                                        say("Mod by zero");
                                }
                        }
			sprintf(tmp, "%lf", dvalue3);
			canon_number(tmp);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '#':
			if (ptr[1] == '#')
			{
				if (stage != NU_ADD)
				{
					ptr = lastop(ptr);
					break;
				}
				else
				{
					*ptr = '\0';
					ptr += 2;
					result1 = next_unit(str, args, arg_flag, stage);
					result2 = next_unit(ptr, args, arg_flag, stage);
					malloc_strcat(&result1, result2);
					new_free(&result2);
					return result1;
				}
			}
			if (ptr[1] == '=') /* #= operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					if (!result1 && !*result1)
						result1 = empty_string;
					malloc_strcat(&result1, result2);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			if (stage != NU_ADD || ptr[1] != '#')
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			malloc_strcat(&result1, result2);
			new_free(&result2);
			return result1;

	/* Reworked - Jeremy Nelson, Feb 1994
	 * & or && should both be supported, each with different
	 * stages, same with || and ^^.  Also, they should be
	 * short-circuit as well.
	 */
		case '&':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				value1 = atol(result1);
				if (value1)
				{
					result2 = next_unit(ptr, args, arg_flag, stage);
					value2 = atol(result2);
					value3 = value1 && value2;
				}
				else
					value3 = 0;
				new_free(&result1);
				new_free(&result2);
                                tmp[0] = '0' + (value3?1:0);
                                tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
			else if (ptr[1] == '=')  /* &= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 &= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 & value2;
                                sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
		case '|':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				value1 = atol(result1);
				if (!value1)
				{
					result2 = next_unit(ptr, args, arg_flag, stage);
					value2 = atol(result2);
					value3 = value1 || value2;
				}
				else
					value3 = 1;
				new_free(&result1);
				new_free(&result2);
				tmp[0] = '0' + (value3 ? 1 : 0);
				tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
			else if (ptr[1] == '=')  /* |= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}


					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 |= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 | value2;
                                sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
		case '^':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				value1 = value1?1:0;
				value2 = value2?1:0;
				value3 = value1 ^ value2;
				new_free(&result1);
				new_free(&result2);
				tmp[0] = '0' + (value3 ? 1 : 0);
				tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
			else if (ptr[1] == '=')  /* ^= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 ^= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 ^ value2;
                                sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
 				return result1;
			}
                case '?':
		{
			/* I forgot to account for nested ?:s... */
			int nested_q = 0;

                        if (stage != NU_TERT)
                        {
				ptr = lastop(ptr);
				break;
			}
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);

			/* Look for the next '?' or ':' */
			ptr2 = sindex(ptr, "?:");

			/* We STOP when we hit a ':' and there isnt
			 * any ? to match it ('cept ourselves...)
			 *
			 * This allows us to continue even if we are
			 * are at nested_q == 0, if the next ?: is a ?.
			 */
			while (ptr2)
			{
				if (*ptr2 == ':')
				{
					if (!nested_q)
						break;
					nested_q--;
				}
				else if (*ptr2 == '?')
					nested_q++;
				ptr2 = sindex(ptr2+1, "?:");
			}
			
			/* Unbalanced :, or possibly missing */
			if (!ptr2)  /* ? but no :, ignore */
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr2++ = '\0';
                        right = result1;
                        value1 = parse_number(&right);
                        if ((value1 == -1) && (*right == (char) 0))
                                value1 = 0;
                        if ( value1 == 0 )
                                while (isspace(*right))
                                        *(right++) = '\0';
                        if ( value1 || *right )
				result2 = next_unit(ptr, args, arg_flag, stage);
			else
				result2 = next_unit(ptr2, args, arg_flag, stage);
                        *(ptr2-1) = ':';
			new_free(&result1);
			return result2;
		}
		case '=':
			if (ptr[1] != '=')
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr++ = '\0';
				result1 = expand_alias((char *) 0, str,
					args, arg_flag, NULL);
				result2 = next_unit(ptr, args, arg_flag, stage);
				display = window_display;
				window_display = 0;
				lastc = result1 + strlen(result1) - 1;
				while (lastc > result1 && *lastc == ' ')
					*lastc-- = '\0';
				for (ptr = result1; *ptr == ' '; ptr++);
				while ((ArrayIndex = (char *) index(ptr, '['))
						!= NULL)
				{
				    *ArrayIndex++='.';
				    if ((EndIndex = MatchingBracket(ArrayIndex,
				        LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
				    {
				        *EndIndex++='\0';
				        strcat(ptr, EndIndex);
				    }
				    else
				        break;
				}
				if (*ptr)
					add_alias(VAR_ALIAS, ptr, result2);
				else
					yell("Invalid assignment expression");
				window_display = display;
				new_free(&result1);
				return result2;
			}
			if (stage != NU_COMP)
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (!my_stricmp(result1, result2))
				malloc_strcpy(&result1, "1");
			else
				malloc_strcpy(&result1, "0");
			new_free(&result2);
			return result1;
		case '>':
		case '<':
			if (ptr[1] == ptr[0])  /* << or >> op */
			{
				/* Not supported yet */
				yell("<< or >> not supported yet");
				break;
			}
			else if (stage != NU_COMP)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			if (ptr[1] == '=')
				value3 = 1, *ptr++ = '\0';
			else
				value3 = 0;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (isdigit(*result1) && isdigit(*result2))
			{
				value1 = atol(result1);
				value2 = atol(result2);
				value1 = (value1 == value2) ? 0 : ((value1 <
					value2) ? -1 : 1);
			}
			else
				value1 = my_stricmp(result1, result2);
			if (value1)
			{
				value2 = (value1 > 0) ? 1 : 0;
				if (op == '<')
					value2 = 1 - value2;
			}
			else
				value2 = value3;
			new_free(&result2);
			sprintf(tmp, "%ld", value2);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '~':
			if (ptr == str)
			{
				if (stage != NU_UNIT)
					break;
				result1 = next_unit(str+1, args, arg_flag, stage);
				if (isdigit(*result1))
				{
					value1 = atol(result1);
					value2 = ~value1;
				}
				else
					value2 = 0;
				sprintf(tmp, "%ld", value2);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
                        else
                        {
                                ptr = lastop(ptr);
                                break;
                        }
		case '!':
			if (ptr == str)
			{
				if (stage != NU_UNIT)
					break;
				result1 = next_unit(str+1, args, arg_flag, stage);

				/* digits can start with -'s, cant they? */
				if (isdigit(*result1) || 
				   ( (*result1 == '-') && isdigit(result1[1]) ))
				{
					value1 = atol(result1);
					value2 = value1 ? 0 : 1;
				}
				else
				{
					value2 = ((*result1)?0:1);
				}
				sprintf(tmp, "%ld", value2);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			if (stage != NU_COMP || ptr[1] != '=')
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (!my_stricmp(result1, result2))
				malloc_strcpy(&result1, "0");
			else
				malloc_strcpy(&result1, "1");
			new_free(&result2);
			return result1;
                case ',': 
			/*
			 * this utterly kludge code is needed (?) to get
			 * around bugs introduced from hop's patches to
			 * alias.c.  the $, variable stopped working
			 * because of this.  -mrg, july 94.
			 */
			if (ptr == str || (ptr > str && ptr[-1] == '$'))
				break;
                        if (stage != NU_EXPR)
                        {
				ptr = lastop(ptr);
				break;
			}
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			new_free(&result1);
			return result2;
		}
	}
	if (stage != NU_UNIT)
		return next_unit(str, args, arg_flag, stage + 1);
	if (isdigit(*str) || *str == '+' || *str == '-')
		malloc_strcpy(&result1, str);
	else
	{
		if (*str == '#' || *str=='@')
			op = *str++;
		else
			op = '\0';
		result1 = find_inline(str);
		if (!result1)
			malloc_strcpy(&result1, empty_string);
		if (op)
		{
			if (op == '#')
				value1 = word_count(result1);
			else if (op == '@')
				value1 = strlen(result1);
			sprintf(tmp, "%ld", value1);
			malloc_strcpy(&result1, tmp);
		}
	}
	return result1;
}
#endif

/*
 * parse_inline:  This evaluates user-variable expression.  I'll talk more
 * about this at some future date. The ^ function and some fixes by
 * troy@cbme.unsw.EDU.AU (Troy Rollo) 
 */
#ifdef __STDC__
char	*parse_inline(char *str, char *args, int *args_flag)
#else
char	*parse_inline(str, args, args_flag)
	char	*str;
	char	*args;
	int	*args_flag;
#endif
{
#ifdef NEW_EVAL
	return evaluate_infix (str, args, args_flag);
#else
	return next_unit(str, args, args_flag, NU_EXPR);
#endif
}

/*
 * arg_number: Returns the argument 'num' from 'str', or, if 'num' is
 * negative, returns from argument 'num' to the end of 'str'.  You might be
 * wondering what's going on down there... here goes.  First we copy 'str' to
 * malloced space.  Then, using next_arg(), we strip out each argument ,
 * putting them in arg_list, and putting their position in the original
 * string in arg_list_pos.  Anyway, once parsing is done, the arguments are
 * returned directly from the arg_list array... or in the case of negative
 * 'num', the arg_list_pos is used to return the postion of the rest of the
 * args in the original string... got it?  Anyway, the bad points of the
 * routine:  1) Always parses out everything, even if only one arg is used.
 * 2) The malloced stuff remains around until arg_number is called with a
 * different string. Even then, some malloced stuff remains around.  This can
 * be fixed. 
 */
/* Think about using the routines in words.c instead of this.
 * Theyre more specialized, and they (as far as i can tell) work more
 * reliably then this monster.
 */
#define	LAST_ARG 8000

#ifdef __STDC__
extern	char	*arg_number(int lower_lim, int upper_lim, char *str)
#else
extern	char	*arg_number(lower_lim, upper_lim, str)
int	lower_lim,
	upper_lim;
char	*str;
#endif
{
	char	*ptr,
		*arg,
		c;
	int	use_full = 0;
	unsigned int	pos,
		start_pos;
	static	char	*last_args = (char *) 0;
	static	char	*last_range = (char *) 0;
	static	char	**arg_list = (char **) 0;
	static	unsigned int	*arg_list_pos = (unsigned int *) 0;
	static	unsigned int	*arg_list_end_pos = (unsigned int *) 0;
	static	int	arg_list_size;

	if (eval_args)
	{
		int	arg_list_limit;

		eval_args = 0;
		new_free((char **)&arg_list);
		new_free((char **)&arg_list_pos);
		new_free((char **)&arg_list_end_pos);
		arg_list_size = 0;
		arg_list_limit = 10;
		arg_list = (char **) new_malloc(sizeof(char *) *
			arg_list_limit);
		arg_list_pos = (unsigned int *) new_malloc(sizeof(unsigned int)
			* arg_list_limit);
		arg_list_end_pos = (unsigned int *) new_malloc(sizeof(unsigned
			int) * arg_list_limit);
		malloc_strcpy(&last_args, str);
		ptr = last_args;
		pos = 0;
		while ((arg = alias_arg(&ptr, &start_pos)) != NULL)
		{
			arg_list_pos[arg_list_size] = pos;
			pos += start_pos + strlen(arg);
			arg_list_end_pos[arg_list_size] = pos++;
			arg_list[arg_list_size++] = arg;
			if (arg_list_size == arg_list_limit)
			{
				arg_list_limit += 10;
				arg_list = (char **)new_realloc((char *)arg_list,
					sizeof(char *) * arg_list_limit);
				arg_list_pos = (unsigned int *)
					new_realloc((char *)arg_list_pos,
					sizeof(unsigned int) * arg_list_limit);
				arg_list_end_pos = (unsigned int *)
					new_realloc((char *)arg_list_end_pos,
					sizeof(unsigned int) *
					arg_list_limit);
			}
		}
	}
	if (upper_lim == LAST_ARG && lower_lim == LAST_ARG)
		upper_lim = lower_lim = arg_list_size - 1;
	if (arg_list_size == 0)
		return (empty_string);
	if ((upper_lim >= arg_list_size) || (upper_lim < 0))
	{
		use_full = 1;
		upper_lim = arg_list_size - 1;
	}
	if (upper_lim < lower_lim)
                return empty_string;
	if (lower_lim >= arg_list_size)
		lower_lim = arg_list_size - 1;
	else if (lower_lim < 0)
		lower_lim = 0;
	if ((use_full == 0) && (lower_lim == upper_lim))
		return (arg_list[lower_lim]);
	c = *(str + arg_list_end_pos[upper_lim]);
	*(str + arg_list_end_pos[upper_lim]) = (char) 0;
	malloc_strcpy(&last_range, str + arg_list_pos[lower_lim]);
	*(str + arg_list_end_pos[upper_lim]) = c;
	return (last_range);
}

/*
 * parse_number: returns the next number found in a string and moves the
 * string pointer beyond that point	in the string.  Here's some examples: 
 *
 * "123harhar"  returns 123 and str as "harhar" 
 *
 * while: 
 *
 * "hoohar"     returns -1  and str as "hoohar" 
 */
#ifdef __STDC__
extern	int	parse_number(char **str)
#else
extern	int	parse_number(str)
	char	**str;
#endif
{
	int	ret;
	char	*ptr;

	ptr = *str;
	if (isdigit(*ptr))
	{
		ret = atoi(ptr);
		for (; isdigit(*ptr); ptr++);
		*str = ptr;
	}
	else
		ret = -1;
	return (ret);
}

void	do_alias_string _((void))
{
	malloc_strcpy(&alias_string, get_input());
	irc_io_loop = 0;
}

/*
 * expander_addition: This handles string width formatting for irc variables
 * when [] is specified.  
 */
#ifdef __STDC__
static	void	expander_addition(char *buff, char *add, int length, char *quote_em)
#else
static	void	expander_addition(buff, add, length, quote_em)
	char	*buff,
		*add;
	int	length;
	char	*quote_em;
#endif
{
	char	format[40],
		*ptr;

	if (length)
	{
		sprintf(format, "%%%d.%ds", -length, (length < 0 ? -length :
			length));
		sprintf(buffer, format, add);
		add = buffer;
	}
	if (quote_em)
	{
		ptr = double_quote(add, quote_em);
		strmcat(buff, ptr, BIG_BUFFER_SIZE);
		new_free(&ptr);
	}
	else
                if (buff)
		        strmcat(buff, add, BIG_BUFFER_SIZE);
}

/* MatchingBracketBackwards returns the matching bracket of the given type */
#ifdef __STDC__
char	*MatchingBracketBackwards(char *string, char *beg, char left, char right)
#else
char	*MatchingBracketBackwards(string, beg, left, right)
	char	*string;
	char	*beg;
	char	left,
		right;
#endif
{
	int	bracket_count = 1;

	while ((string > beg) && bracket_count)
	{
		if (*string == left)
			bracket_count++;
		else if (*string == right)
		{
			if (!--bracket_count)
				return string;
		}
		else if (string[-1] == '\\')
			string--;
		string--;
	}
	return (char *) 0;
}

/* MatchingBracket returns the next unescaped bracket of the given type */
#ifdef __STDC__
char	*MatchingBracket(char *string, char left, char right)
#else
char	*MatchingBracket(string, left, right)
	char	*string;
	char	left,
		right;
#endif
{
	int	bracket_count = 1;

	while (*string && bracket_count)
	{
		if (*string == left)
			bracket_count++;
		else if (*string == right)
		{
			if (!--bracket_count)
				return string;
		}
		else if (*string == '\\' && string[1])
			string++;
		string++;
	}
	return (char *) 0;
}


/*
 * alias_special_char: Here we determin what to do with the character after
 * the $ in a line of text. The special characters are described more fulling
 * in the help/ALIAS file.  But they are all handled here. Paremeters are the
 * name of the alias (if applicable) to prevent deadly recursion, a
 * destination buffer (of size BIG_BUFFER_SIZE) to which things are appended,
 * a ptr to the string (the first character of which is the special
 * character, the args to the alias, and a character indication what
 * characters in the string should be quoted with a backslash.  It returns a
 * pointer to the character right after the converted alias.

 The args_flag is set to 1 if any of the $n, $n-, $n-m, $-m, $*, or $() is used
 in the alias.  Otherwise it is left unchanged.
 */
/*ARGSUSED*/
#ifdef __STDC__
static	char	*alias_special_char(char *name, char *buffer, char *ptr, char *args, char *quote_em, int *args_flag)
#else
static	char	*alias_special_char(name, buffer, ptr, args, quote_em,args_flag)
	char	*name;
	char	*buffer;
	char	*ptr;
	char	*args;
	char	*quote_em;
	int	*args_flag;
#endif
{
	char	*tmp,
		c;
	int	upper,
	lower,
	length;

	length = 0;
	if ((c = *ptr) == LEFT_BRACKET)
	{
		ptr++;
		if ((tmp = (char *) index(ptr, RIGHT_BRACKET)) != NULL)
		{
			*(tmp++) = (char) 0;
			length = atoi(ptr);
			ptr = tmp;
			c = *ptr;
		}
		else
		{
			say("Missing %c", RIGHT_BRACKET);
			return (ptr);
		}
	}
	tmp = ptr+1;
	switch (c)
	{
	case LEFT_PAREN:
		{
			char	sub_buffer[BIG_BUFFER_SIZE+1];

			if ((ptr = MatchingBracket(tmp, LEFT_PAREN,
			    RIGHT_PAREN)) || (ptr = (char *) index(tmp,
			    RIGHT_PAREN)))
				*(ptr++) = (char) 0;
			tmp = expand_alias((char *) 0, tmp, args, args_flag,
				NULL);
			*sub_buffer = (char) 0;
			alias_special_char((char *) 0, sub_buffer, tmp,
				args, quote_em,args_flag);
			expander_addition(buffer, sub_buffer, length, quote_em);
			new_free(&tmp);
			*args_flag = 1;
		}
		return (ptr);
	case '!':
		if ((ptr = (char *) index(tmp, '!')) != NULL)
			*(ptr++) = (char) 0;
		if ((tmp = do_history(tmp, empty_string)) != NULL)
		{
			expander_addition(buffer, tmp, length, quote_em);
			new_free(&tmp);
		}
		return (ptr);
	case LEFT_BRACE:
		/* BLAH. This didnt allow for nesting before.  How lame. */
		ptr = MatchingBracket(tmp, LEFT_BRACE, RIGHT_BRACE);
		if (!ptr)
		{
			yell("Unmatched %c", LEFT_BRACE);
			break;
		}
		*(ptr++) = (char) 0;
		if ((tmp = parse_inline(tmp, args, args_flag)) != NULL)
		{
			expander_addition(buffer, tmp, length, quote_em);
			new_free(&tmp);
		}
		return (ptr);
	case DOUBLE_QUOTE:
		if ((ptr = (char *) index(tmp, DOUBLE_QUOTE)) != NULL)
			*(ptr++) = (char) 0;
		alias_string = (char *) 0;
		if (irc_io(tmp, do_alias_string, use_input, 1))
		{
			yell("Illegal recursive edit");
			break;
		}
		expander_addition(buffer, alias_string, length, quote_em);
		new_free(&alias_string);
		return (ptr);
	case '*':
		expander_addition(buffer, args, length, quote_em);
		*args_flag = 1;
		return (ptr + 1);
	default:
		if (isdigit(c) || (c == '-') || c == '~')
		{
			*args_flag = 1;
			if (*ptr == '~')
			{
				lower = upper = LAST_ARG;
				ptr++;
			}
			else
			{
				lower = parse_number(&ptr);
				if (*ptr == '-')
				{
					ptr++;
					upper = parse_number(&ptr);
				}
				else
					upper = lower;
			}
			expander_addition(buffer, arg_number(lower, upper,
				args), length, quote_em);
			return (ptr ? ptr : empty_string);
		}
		else
		{
			char	*rest,
				c = (char) 0;

		/*
		 * Why use ptr+1?  Cause try to maintain backward compatability
		 * can be a pain in the butt.  Basically, we don't want any of
		 * the illegal characters in the alias, except that things like
		 * $* and $, were around first, so they must remain legal.  So
		 * we skip the first char after the $.  Does this make sense?
		 */
			/* special case for $ */
			if (*ptr == '$')
			{
				rest = ptr+1;
				c = *rest;
				*rest = (char) 0;
			}
			else if ((rest = sindex(ptr+1, alias_illegals)) != NULL)
			{
				if (isalpha(*ptr) || *ptr == '_')
					while ((*rest == LEFT_BRACKET ||
					    *rest == LEFT_PAREN) &&
					    (tmp = MatchingBracket(rest+1,
					    *rest, (*rest == LEFT_BRACKET) ?
					    RIGHT_BRACKET: RIGHT_PAREN)))
						rest = tmp + 1;
				c = *rest;
				*rest = (char) 0;
			}
			if ((tmp = parse_inline(ptr, args, args_flag)) != NULL)
			{
				expander_addition(buffer, tmp, length,
					quote_em);
				new_free(&tmp);
			}
			if (rest)
				*rest = c;
			return(rest);
		}
	}
	return NULL;
}


/*
 * expand_alias: Expands inline variables in the given string and returns the
 * expanded string in a new string which is malloced by expand_alias(). 
 *
 * Also unescapes anything that was quoted with a backslash
 *
 * Behaviour is modified by the following:
 *	Anything between brackets (...) {...} is left unmodified.
 *	If more_text is supplied, the text is broken up at
 *		semi-colons and returned one at a time. The unprocessed
 *		portion is written back into more_text.
 *	Backslash escapes are unescaped.
 */
#ifdef __STDC__
char	*expand_alias(char *name, char *string, char *args, int *args_flag, char **more_text)
#else
char	*expand_alias(name, string, args,args_flag, more_text)
	char	*name,
		*string,
		*args;
	int	*args_flag;
	char	**more_text;
#endif
{
	char	buffer[BIG_BUFFER_SIZE + 1],
		*ptr,
		*stuff = (char *) 0,
		*free_stuff;
	char	*quote_em,
		*quote_str = (char *) 0;
	char	ch;
	int	quote_cnt = 0;
	int	is_quote = 0;
	void	(*str_cat) _((char *, char *, int));

	/* This is a kludge i put in here until i can figure
	   out why this is recursive */
	/* this *might* be a memory leak... just keep an eye on it */
	if (!string || !*string)
		return m_strdup(empty_string);

	if (*string == '@' && more_text)
	{
		str_cat = strmcat;
		*args_flag = 1; /* Stop the @ command from auto appending */
	}
	else
		str_cat = strmcat_ue;
	malloc_strcpy(&stuff, string);
	free_stuff = stuff;
	*buffer = (char) 0;
	eval_args = 1;
	ptr = stuff;
	if (more_text)
		*more_text = NULL;
	while (ptr && *ptr)
	{
		if (is_quote)
		{
			is_quote = 0;
			++ptr;
			continue;
		}
		switch(*ptr)
		{
		case '$':
	/*
	 * The test here ensures that if we are in the expression
	 * evaluation command, we don't expand $. In this case we
	 * are only coming here to do command separation at ';'s.
	 * If more_text is not defined, and the first character is
	 * '@', we have come here from [] in an expression.
	 */
			if (more_text && *string == '@')
			{
				ptr++;
				break;
			}
			*(ptr++) = (char) 0;
			(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
			while (*ptr == '^')
			{
				ptr++;
				if (quote_str)
					quote_str = (char *)
						new_realloc((char *)quote_str,
						sizeof(char) * (quote_cnt + 2));
				else
					quote_str = (char *)
						new_malloc(sizeof(char) *
						(quote_cnt + 2));
				quote_str[quote_cnt++] = *(ptr++);
				quote_str[quote_cnt] = (char) 0;
			}
			quote_em = quote_str;
			/* MEMORY LEAK */
			stuff = alias_special_char(name, buffer, ptr, args,
				quote_em, args_flag);
			if (stuff)
				new_free(&quote_str);
			quote_cnt = 0;
			ptr = stuff;
			break;
		case ';':
			if (!more_text)
			{
				ptr++;
				break;
			}
			*more_text = string + (ptr - free_stuff) +1;
			*ptr = '\0'; /* To terminate the loop */
			break;
		case LEFT_PAREN:
		case LEFT_BRACE:
			ch = *ptr;
			*ptr = '\0';
			(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
			stuff = ptr;
			*args_flag = 1;
			if (!(ptr = MatchingBracket(stuff + 1, ch,
					(ch == LEFT_PAREN) ?
					RIGHT_PAREN : RIGHT_BRACE)))
			{
				yell("Unmatched %c", ch);
				ptr = stuff + strlen(stuff+1)+1;
			}
			else
				ptr++;
			*stuff = ch;
			ch = *ptr;
			*ptr = '\0';
			strmcat(buffer, stuff, BIG_BUFFER_SIZE);
			stuff = ptr;
			*ptr = ch;
			break;
		case '\\':
			is_quote = 1;
			ptr++;
			break;
		default:
			ptr++;
			break;
		}
	}
	if (stuff)
		(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
	ptr = (char *) 0;
	new_free(&free_stuff);
	malloc_strcpy(&ptr, buffer);
	if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
		yell("Expanded [%s] to [%s]",
			string, ptr);
	return (ptr);
}

/*
 * get_alias: returns the alias matching 'name' as the function value. 'args'
 * are expanded as needed, etc.  If no matching alias is found, null is
 * returned, cnt is 0, and full_name is null.  If one matching alias is
 * found, it is retuned, with cnt set to 1 and full_name set to the full name
 * of the alias.  If more than 1 match are found, null is returned, cnt is
 * set to the number of matches, and fullname is null. NOTE: get_alias()
 * mallocs the space for the full_name, but returns the actual value of the
 * alias if found! 
 */
#ifdef __STDC__
char	*get_alias(int type, char *name, int *cnt, char **full_name)
#else
char	*get_alias(type, name, cnt, full_name)
	int	type;
	char	*name,
		**full_name;
	int	*cnt;
#endif
{
	Alias	*tmp;

	*full_name = (char *) 0;
	if ((name == (char *) 0) || (*name == (char) 0))
	{
		*cnt = 0;
		return ((char *) 0);
	}
	if ((tmp = find_alias(&(alias_list[type]), name, 0, cnt)) != NULL)
	{
		if (*cnt < 2)
		{
			malloc_strcpy(full_name, tmp->name);
			return (tmp->stuff);
		}
	}
	return ((char *) 0);
}

/*
 * match_alias: this returns a list of alias names that match the given name.
 * This is used for command completion etc.  Note that the returned array is
 * malloced in this routine.  Returns null if no matches are found 
 */
#ifdef __STDC__
char	**match_alias(char *name, int *cnt, int type)
#else
char	**match_alias(name, cnt, type)
	char	*name;
	int	*cnt;
	int	type;
#endif
{
	Alias	*tmp;
	char	**matches = (char **) 0;
	int	matches_size = 5;
	int	len;
	char	*last_match = (char *) 0;
	char	*dot;

	len = strlen(name);
	*cnt = 0;
	matches = (char	**) new_malloc(sizeof(char *) * matches_size);
	for (tmp = alias_list[type]; tmp; tmp = tmp->next)
	{
		if (strncmp(name, tmp->name, len) == 0)
		{
			if ((dot = (char *) index(tmp->name+len, '.')) != NULL)
			{
				if (type == COMMAND_ALIAS)
					continue;
				else
				{
					*dot = '\0';
					if (last_match && !strcmp(last_match,
							tmp->name))
					{
						*dot = '.';
						continue;
					}
				}
			}
			matches[*cnt] = (char *) 0;
			malloc_strcpy(&(matches[*cnt]), tmp->name);
			last_match = matches[*cnt];
			if (dot)
				*dot = '.';
			if (++(*cnt) == matches_size)
			{
				matches_size += 5;
				matches = (char	**) new_realloc((char *)matches,
					sizeof(char *) * matches_size);
			}
		}
		else if (*cnt)
			break;
	}
	if (*cnt)
	{
		matches = (char	**) new_realloc((char *)matches, sizeof(char *) *
			(*cnt + 1));
		matches[*cnt] = (char *) 0;
	}
	else
		new_free((char **)&matches);
	return (matches);
}

/* delete_alias: The alias name is removed from the alias list. */
#ifdef __STDC__
void delete_alias(int type, char *name)
#else
void delete_alias(type, name)
	int	type;
	char	*name;
#endif
{
	Alias	*tmp;

	upper(name);
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) NULL))
			!= NULL)
	{
		new_free(&(tmp->name));
		new_free(&(tmp->stuff));
		new_free((char **)&tmp);
		if (type == COMMAND_ALIAS)
			say("Alias	%s removed", name);
		else
			say("Assign %s removed", name);
	}
	else
		say("No such alias: %s", name);
}

/*
 * list_aliases: Lists all aliases matching 'name'.  If name is null, all
 * aliases are listed 
 */
#ifdef __STDC__
void list_aliases(int type, char *name)
#else
void list_aliases(type, name)
	int	type;
	char	*name;
#endif
{
	Alias	*tmp;
	int	len;
	int	DotLoc,
		LastDotLoc = 0;
	char	*LastStructName = NULL;
	char	*s;

	if (type == COMMAND_ALIAS)
		say("Aliases:");
	else
		say("Assigns:");
	if (name)
	{
		upper(name);
		len = strlen(name);
	}
	else
		len = 0;
	for (tmp = alias_list[type]; tmp; tmp = tmp->next)
	{
		if (!name || !strncmp(tmp->name, name, len))
		{
			s = index(tmp->name + len, '.');
			if (!s)
			{
				if (tmp->stub)
					say("\t%s STUBBED TO %s", tmp->name, tmp->stub);
				else
					say("\t%s\t%s", tmp->name, tmp->stuff);
			}
			else
			{
				DotLoc = s - tmp->name;
				if (!LastStructName || (DotLoc != LastDotLoc) || strncmp(tmp->name, LastStructName, DotLoc))
				{
					say("\t%*.*s\t<Structure>", DotLoc, DotLoc, tmp->name);
					LastStructName = tmp->name;
					LastDotLoc = DotLoc;
				}
			}
		}
	}
}

/*
 * mark_alias: sets the mark field of the given alias to 'flag', and returns
 * the previous value of the mark.  If the name is not found, -1 is returned.
 * This is used to prevent recursive aliases by marking and unmarking
 * aliases, and not reusing an alias that has previously been marked.  I'll
 * explain later 
 */
#ifdef __STDC__
int mark_alias(char *name, int flag)
#else
int mark_alias(name, flag)
	char	*name;
	int	flag;
#endif
{
	int	old_mark;
	Alias	*tmp;
	int	match;

	if ((tmp = find_alias(&(alias_list[COMMAND_ALIAS]), name, 0, &match))
			!= NULL)
	{
		if (match < 2)
		{
			old_mark = tmp->mark;
		/* New handling of recursion */
			if (flag)
			{
				int	i;
				/* Count recursion */

				tmp->mark = tmp->mark + flag;
				if ((i = get_int_var(MAX_RECURSIONS_VAR)) > 1)
				{
					if (tmp->mark > i)
					{
						tmp->mark = 0;
						return(1); /* MAX exceeded. */
					}
					else return(0);
				/* In recursion but it's ok */
				}
				else
				{
					if (tmp->mark > 1)
					{
						tmp->mark = 0;
						return(1);
				/* max of 1 here.. exceeded */
					}
					else return(0);
				/* In recursion but it's ok */
				}
			}
			else
		/* Not in recursion at all */
			{
				tmp->mark = 0;
				return(old_mark);
			/* This one gets ignored anyway */
			}
		}
	}
	return (-1);
}

/*
 * execute_alias: After an alias has been identified and expanded, it is sent
 * here for proper execution.  This routine mainly prevents recursive
 * aliasing.  The name is the full name of the alias, and the alias is
 * already expanded alias (both of these parameters are returned by
 * get_alias()) 
 */
#ifdef __STDC__
void execute_alias(char *alias_name, char *alias, char *args)
#else
void execute_alias(alias_name, alias, args)
	char	*alias_name,
		*alias,
		*args;
#endif
{
	if (mark_alias(alias_name, 1))
		say("Maximum recursion count exceeded in: %s", alias_name);
	else
	{
		parse_line(alias_name, alias, args, 0,1);
		mark_alias(alias_name, 0);
	}
}

/*
 * save_aliases: This will write all of the aliases to the FILE pointer fp in
 * such a way that they can be read back in using LOAD or the -l switch 
 */
#ifdef __STDC__
void save_aliases(FILE *fp, int do_all)
#else
void save_aliases(fp, do_all)
	FILE	*fp;
	int	do_all;
#endif
{
	Alias	*tmp;

	for (tmp = alias_list[VAR_ALIAS]; tmp; tmp = tmp->next)
	{
		if (!tmp->global || do_all)
		{
			if (tmp->stub)
				fprintf(fp, "STUB ASSIGN %s %s\n", tmp->name, tmp->stub);
			else
				fprintf(fp, "ASSIGN %s %s\n", tmp->name, tmp->stuff);
		}
	}
	for (tmp = alias_list[COMMAND_ALIAS]; tmp; tmp = tmp->next)
	{
		if (!tmp->global || do_all)
		{
			if (tmp->stub)
				fprintf(fp, "STUB ASSIGN %s %s\n", tmp->name, tmp->stub);
			else
				fprintf(fp, "ALIAS %s %s\n", tmp->name, tmp->stuff);
		}
	}	
}

static void flush_variables _((void));
static void flush_aliases2 _((void));
extern void flush_on_hooks _((void));

#ifdef __STDC__
extern  void	flush_aliases(char *command,char *args, char *subargs)
#else
extern  void	flush_aliases(command,args,subargs)
	char 	*command,
		*args;
	char	*subargs;
#endif
{
        char 	*blah;

	if (!args || !*args)
	{
		flush_aliases2();
		flush_variables();
		flush_on_hooks();
		return;
	}
        while ((blah = next_arg(args, &args)))
        {
		if (!my_strnicmp(blah,"VAR",1))
			flush_variables();
		else if (!my_strnicmp(blah,"ALIAS",3))
			flush_aliases2();
		else if (!my_strnicmp(blah,"ON",1))
			flush_on_hooks();
		else if (!my_strnicmp(blah,"ALL",3))
		{
			flush_aliases2();
			flush_variables();
			flush_on_hooks();
		}
	}
        yell("*** I hope you didnt have anything valuable in there...");
}

void flush_aliases2 _((void))
{
	Alias	*tmp;
        Alias   *ntmp;

	for (tmp = alias_list[COMMAND_ALIAS]; tmp;)
	{
		tmp = find_alias(&alias_list[COMMAND_ALIAS],tmp->name,1,NULL);
		new_free(&tmp->stuff);
		new_free(&tmp->name);
		ntmp = tmp->next;
		new_free((char **)&tmp);
		tmp = ntmp;
	}
	alias_list[COMMAND_ALIAS] = (Alias *) 0;
}

void flush_variables _((void))
{
	Alias	*tmp;
        Alias   *ntmp;

	for (tmp = alias_list[VAR_ALIAS]; tmp;)
	{
		tmp = find_alias(&alias_list[VAR_ALIAS],tmp->name,1,NULL);
                new_free(&tmp->stuff);
                new_free(&tmp->name);
                ntmp = tmp->next;
                new_free((char **)&tmp);
                tmp = ntmp;
	}
        alias_list[VAR_ALIAS] = (Alias *) 0;
}

/* built in expando functions */
static	char	*
alias_line _((void))
{
	return(get_input());
}

static	char	*
alias_buffer _((void))
{
	return cut_buffer;
}

static	char	*
alias_time _((void))
{
	return(update_clock(GET_TIME));
}

static	char	*
alias_dollar _((void))
{
	return("$");
}

static	char	*
alias_detected _((void))
{
	return last_notify_nick;
}

static	char	*
alias_nick _((void))
{
	return(get_server_nickname(curr_scr_win->server));
}

static	char	*
alias_away _((void))
{
	return server_list[curr_scr_win->server].away;
}

static	char	*
alias_sent_nick _((void))
{
	return (sent_nick) ? sent_nick : empty_string;
}

static	char	*
alias_recv_nick _((void))
{
	return (recv_nick) ? recv_nick : empty_string;
}

static	char	*
alias_msg_body _((void))
{
	return (sent_body) ? sent_body : empty_string;
}

static	char	*
alias_joined_nick _((void))
{
	return (joined_nick) ? joined_nick : empty_string;
}

static	char	*
alias_public_nick _((void))
{
	return (public_nick) ? public_nick : empty_string;
}

static	char	*
alias_channel _((void))
{
	char	*tmp;

	return (tmp = get_channel_by_refnum(0)) ? tmp : "0";
}

static	char	*
alias_server _((void))
{
	return (parsing_server_index == -1) ?
		get_server_itsname(parsing_server_index) :
		(get_window_server(0) != -1) ?
			get_server_itsname(get_window_server(0)) : empty_string;
}

static	char	*
alias_query_nick _((void))
{
	char	*tmp;

	return (tmp = query_nick()) ? tmp : empty_string;
}

static	char	*
alias_target _((void))
{
	char	*tmp;

	return (tmp = get_target_by_refnum(0)) ? tmp : empty_string;
}

static	char	*
alias_invite _((void))
{
	return (invite_channel) ? invite_channel : empty_string;
}

static	char	*
alias_cmdchar _((void))
{
	static	char	thing[2];
	char	*cmdchars;

	if ((cmdchars = get_string_var(CMDCHARS_VAR)) == (char *) 0)
		cmdchars = DEFAULT_CMDCHARS;
	thing[0] = cmdchars[0];
	thing[1] = (char) 0;
	return(thing);
}

static	char	*
alias_oper _((void))
{
	return get_server_operator(from_server) ?
		get_string_var(STATUS_OPER_VAR) : empty_string;
}

static	char	*
alias_chanop _((void))
{
	char	*tmp;

	return ((tmp = get_channel_by_refnum(0)) && get_channel_oper(tmp,
			from_server)) ?
		"@" : empty_string;
}

static	char	*
alias_modes _((void))
{
	char	*tmp;

	return (tmp = get_channel_by_refnum(0)) ?
		get_channel_mode(tmp, from_server) : empty_string;
}

extern time_t start_time;

static  char    *
alias_online _((void))
{
        static char booya[100];
        sprintf(booya, "%ld",(long)start_time);
        return booya;
}

static  char    *
alias_idle _((void))
{ 
        static char booya[100];
        sprintf(booya, "%ld",time((time_t *)0)-idle_time);
        return booya;
}

static  char    *
alias_show_userhost _((void))
{
        static char    booya[100];
        
        sprintf(booya, "%s@%s",username, hostname);
        return booya;
}

static  char    *
alias_show_realname _((void))
{
        return realname;
}

static	char	*
alias_version_str _((void))
{
	return irc_version;
}

static	char	*alias_version _((void))
{
	char *result = (char *) 0;
	malloc_strcpy(&result, internal_version);
	return result;
}

static	char	*alias_currdir  _((void))
{
	char	*result = (char *) 0;

        getcwd(buffer, BIG_BUFFER_SIZE+1);
	/* is this more portable? *shrug*. i chose BIG_BUFFER_SIZE+1
	   because it's more readable -cgw- */
	/* getcwd((char *)buffer, sizeof((char *)buffer)); */
	malloc_strcpy(&result, buffer);
	return result;
}

static	char	*alias_current_numeric  _((void))
{
	char	*result = (char *) 0,
		number[4];
	
	sprintf(number, "%03d", -current_numeric);
	malloc_strcpy(&result, number);
	return result;
}

static	char	*alias_server_version  _((void))
{
	char	*result = (char *) 0,
		*s;

	malloc_strcpy(&result,
		(s = server_list[curr_scr_win->server].version_string) ?
				s : empty_string);
	return result;
}

/*
 * alias: the /ALIAS command.  Calls the correct alias function depending on
 * the args 
 */
#ifdef __STDC__
void	alias(char *command, char *args, char *subargs)
#else
void	alias(command, args, subargs)
	char	*command,
		*args;
	char	*subargs;
#endif
{
	char	*name,
		*rest;
	int	type;
	char	*ArrayIndex;
	char	*EndIndex;

	type = *command - 48;	/*
				 * A trick!  Yes, well, what the hell.  Note
                                 * the the command part of ALIAS is "0" and
                                 * the command part of ASSIGN is "1" in the
				 * command array list
				 */
	if ((name = next_arg(args, &rest)) != NULL)
	{
		while ((ArrayIndex = (char *) index(name, '[')) != NULL)
		{
			*ArrayIndex++ = '.';
			if ((EndIndex = MatchingBracket(ArrayIndex,
					LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
			{
				*EndIndex++ = '\0';
				strcat(name, EndIndex);
			}
			else
				break;
		}
		if (*rest)
		{
			if (*rest == LEFT_BRACE)
			{
				char	*ptr = MatchingBracket(++rest,
						LEFT_BRACE, RIGHT_BRACE);
				if (!ptr)
				    say("Unmatched brace in ALIAS or ASSIGN");
				else if (ptr[1])
				{
					say("Junk after closing brace in ALIAS or ASSIGN");
				}
				else
				{
					*ptr = '\0';
					add_alias(type, name, rest);
				}
			}
			else
				add_alias(type, name, rest);
		}
		else
		{
			if (*name == '-')
			{
				if (*(name + 1))
					delete_alias(type, name + 1);
				else
					say("You must specify an alias to be removed");
			}
			else
				list_aliases(type, name);
		}
	}
	else
		list_aliases(type, (char *) 0);
}

/* stub command -- makes a stub for an alias/assign */
#ifdef __STDC__
void stubcmd (char *command, char *args, char *subargs)
#else
void stubcmd (command, args, subargs)
char *command;
char *args;
char *subargs;
#endif
{
	char	*name,
		*rest;
	char	*type_c;
	int	type;
	char	*ArrayIndex;
	char	*EndIndex;

	type_c = next_arg(args, &rest);
	if ((type_c == NULL) || (*type_c == 0))
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}
	else
	{
		if (!my_strnicmp(type_c, "ASSIGN", 6))
			type = 1;
		else if (!my_strnicmp(type_c, "ALIAS", 5))
			type = 0;
		else
		{
			yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
			return;
		}
	}

	name = next_arg(rest, &rest);
	if ((name == NULL) || (*name == 0))
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}
	else
	{
		while ((ArrayIndex = (char *) index(name, '[')) != NULL)
		{
			*ArrayIndex++ = '.';
			if ((EndIndex = MatchingBracket(ArrayIndex,
					LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
			{
				*EndIndex++ = '\0';
				strcat(name, EndIndex);
			}
			else
				break;
		}
	}
	if ((rest == NULL) || (*rest == 0))
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}

	add_stub(type, name, rest);
}

#ifdef __STDC__
char	* function_left(char *input)
#else
char	* function_left(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*count;
	int	cvalue;

	count = new_next_arg(input, &input);
	if (count)
		cvalue = atoi(count);
	else
		cvalue = 0;
	if (strlen(input) > cvalue)
		input[cvalue] = '\0';
	malloc_strcpy(&result, input);
	return result;
}

#ifdef __STDC__
char	*function_right(char *input)
#else
char	*function_right(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*count;
	int	cvalue;

	count = new_next_arg(input, &input);
	if (count)
		cvalue = atoi(count);
	else
		cvalue = 0;
	if (strlen(input) > cvalue)
		input += strlen(input) - cvalue;
	malloc_strcpy(&result, input);
	return result;
}

#ifdef __STDC__
char	*function_mid(char *input)
#else
char	*function_mid(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*mid_index;
	int	ivalue;
	char	*count;
	int	cvalue;

	mid_index = new_next_arg(input, &input);
	if (mid_index)
		ivalue = atoi(mid_index);
	else
		ivalue = 0;
	count = new_next_arg(input, &input);
	if (count)
		cvalue = atoi(count);
	else
		cvalue = 0;
	if (strlen(input) > ivalue)
		input += ivalue;
	else
		*input = (char) 0;
	if (strlen(input) > cvalue)
		input[cvalue] = (char) 0;
	malloc_strcpy(&result, input);
	return result;
}

/* patch from Sarayan to make $rand() better */

#define RAND_A 16807L
#define RAND_M 2147483647L
#define RAND_Q 127773L
#define RAND_R 2836L

#ifdef __STDC__
static	long	randm(long l)
#else
static	long	randm(l)
	long	l;
#endif
{
	static	u_long	z = 0;
	long	t;

	if (!z)
		z = (u_long) getuid();
	if (!l)
	{
		t = RAND_A * (z % RAND_Q) - RAND_R * (z / RAND_Q);
		if (t > 0)
			z = t;
		else
			z = t + RAND_M;
		return (z >> 8) | ((z & 255) << 23);
	}
	else
	{
		if (l < 0)
			z = (u_long) getuid();
		else
			z = l;
		return 0;
	}
}

#ifdef __STDC__
char	*function_rand(char *input)
#else
char	*function_rand(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	tmp[40];
	long	tempin;

	sprintf(tmp, "%ld", (tempin = atol(input)) ? randm(0L) % tempin : 0);
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_srand(char *input)
#else
char	*function_srand(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	if (input && *input)
		(void) randm(atol(input));
	else
		(void) randm((time_t) time(NULL));
	malloc_strcpy(&result, empty_string);
	return result;
}

/*ARGSUSED*/
#ifdef __STDC__
char	*function_time(char *input)
#else
char	*function_time(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	time_t	ltime;
	char	tmp[40];

	(void) time(&ltime);
	sprintf(tmp, "%ld", ltime);
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_stime (char *input)
#else
char	* function_stime(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	time_t	ltime;

	ltime = atol(input);
	malloc_strcpy(&result, ctime(&ltime));
	result[strlen(result) - 1] = (char) 0;
	return result;
}

#ifdef __STDC__
char	*function_tdiff (char *input)
#else
char	* function_tdiff(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	time_t	ltime;
	time_t	days,
		hours,
		minutes,
		seconds;
	char	tmp[80];
	char	*tstr;

	ltime = atol(input);
	seconds = ltime % 60;
	ltime = (ltime - seconds) / 60;
	minutes = ltime%60;
	ltime = (ltime - minutes) / 60;
	hours = ltime % 24;
	days = (ltime - hours) / 24;
	tstr = tmp;
	if (days)
	{
		sprintf(tstr, "%ld day%s ", days, (days==1)?"":"s");
		tstr += strlen(tstr);
	}
	if (hours)
	{
		sprintf(tstr, "%ld hour%s ", hours, (hours==1)?"":"s");
		tstr += strlen(tstr);
	}
	if (minutes)
	{
		sprintf(tstr, "%ld minute%s ", minutes, (minutes==1)?"":"s");
		tstr += strlen(tstr);
	}
	if (seconds || (!days && !hours && !minutes))
	{
		sprintf(tstr, "%ld second%s", seconds, (seconds==1)?"":"s");
		tstr += strlen(tstr);
	}
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_index (char *input)
#else
char	* function_index(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*schars;
	char	*iloc;
	int	ival;
	char	tmp[40];

	schars = new_next_arg(input, &input);
	iloc = (schars) ? sindex(input, schars) : NULL;
	ival = (iloc) ? iloc - input : -1;
	sprintf(tmp, "%d", ival);
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_rindex (char *input)
#else
char	* function_rindex(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*schars;
	char	*iloc, *liloc;
	int	ival;
	char	tmp[40];

	schars = new_next_arg(input, &input);
	iloc = NULL;
	if (schars)
	{
		liloc = sindex(input, schars);
		while (liloc)
			liloc = sindex((iloc = liloc) + 1, schars);
	}
	ival = (iloc) ? iloc-input : -1;
	sprintf(tmp, "%d", ival);
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_match (char *input)
#else
char	* function_match(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*pattern;
	char	*word;
	int	current_match;
	int	best_match = 0;
	int	match = 0;
	int	match_index = 0;
	char	tmp[40];

	if ((pattern = new_next_arg(input, &input)) != NULL)
	{
		while ((word = new_next_arg(input, &input)) != NULL)
		{
			match_index++;
			if ((current_match = wild_match(pattern, word))
					> best_match)
			{
				match = match_index;
				best_match=current_match;
			}
		}
	}
	sprintf(tmp, "%d", match);
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_rmatch (char *input)
#else
char	* function_rmatch(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*pattern;
	char	*word;
	int	current_match;
	int	best_match = 0;
	int	match = 0;
	int	rmatch_index = 0;
	char	tmp[40];

	if ((pattern = new_next_arg(input, &input)) != NULL)
	{
		while ((word = new_next_arg(input, &input)) != NULL)
		{
			rmatch_index++;
			if ((current_match = wild_match(word, pattern)) >
					best_match)
			{
				match = rmatch_index;
				best_match = current_match;
			}
		}
	}
	sprintf(tmp, "%d", match);
	malloc_strcpy(&result, tmp);
	return result;
}

/*ARGSUSED*/
#ifdef __STDC__
char	*function_userhost (char *input)
#else
char	* function_userhost(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	malloc_strcpy(&result, FromUserHost ? FromUserHost : empty_string);
	return result;
}

#ifdef __STDC__
char	*function_strip (char *input)
#else
char	* function_strip(input)
	char	*input;
#endif
{
	static	char	result[BIG_BUFFER_SIZE+1];
	char	*retval = (char *) 0;
	char	*chars;
	char	*cp, *dp;

	if ((chars = new_next_arg(input, &input)) && input)
	{
		for (cp = input, dp = result; *cp; cp++)
		{
			if (!index(chars, *cp))
				*dp++ = *cp;
		}
		*dp = '\0';
	}
	malloc_strcpy(&retval, result);
	return retval;
}

#ifdef __STDC__
char	*function_encode (char *input)
#else
char	* function_encode(input)
	unsigned char	*input;
#endif
{
	static	char	result[BIG_BUFFER_SIZE+1];
	char	*retval = (char *) 0;
	unsigned char	*c;
	int	i = 0;

	c = input;
	while (*c)
	{
		result[i++] = (*c >> 4) + 0x41;
		result[i++] = (*c & 0x0f) + 0x41;
		c++;
	}
	result[i] = '\0';
	malloc_strcpy(&retval, result);
	return retval;
}


#ifdef __STDC__
char	*function_decode (char *input)
#else
char	* function_decode(input)
	unsigned char	*input;
#endif
{
	static	char	result[BIG_BUFFER_SIZE+1];
	char	*retval = (char *) 0;
	unsigned char	*c;
	unsigned char	d,e;
	int	i = 0;

	c = input;
	while((d = *c) && (e = *(c+1)))
	{
		result[i] = ((d - 0x41) << 4) | (e - 0x41);
		c += 2;
		i++;
	}
	result[i] = '\0';
	malloc_strcpy(&retval, result);
	return retval;
}

#ifdef __STDC__
char	*function_ischannel (char *input)
#else
char	* function_ischannel(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	malloc_strcpy(&result, is_channel(input) ? "1" : "0");
	return result;
}

#ifdef __STDC__
char	*function_ischanop (char *input)
#else
char	* function_ischanop(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*nick;
	char	*channel = NULL;

	if (!(nick = new_next_arg(input, &channel)))
		malloc_strcpy(&result, "0");
	else
		malloc_strcpy(&result, is_chanop(channel, nick) ? "1" : "0");
	return result;
}


#ifdef __STDC__
char	*function_word (char *input)
#else
char	* function_word(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*count;
	int	cvalue;
	char	*word;

	count = new_next_arg(input, &input);
	if (count)
		cvalue = atoi(count);
	else
		cvalue = 0;
	if (cvalue < 0)
		malloc_strcpy(&result, empty_string);
	else
	{
		for (word = new_next_arg(input, &input); word && cvalue--;
				word = new_next_arg(input, &input))
			;
		malloc_strcpy(&result, (word) ? word : empty_string);
	}
	return result;
}


#ifdef __STDC__
char	*function_winnum (char *input)
#else
char	* function_winnum(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	tmp[10];

	if (curr_scr_win)
		sprintf(tmp, "%d", curr_scr_win->refnum);
	else
		strcpy(tmp, "-1");
	malloc_strcpy(&result, tmp);
	return result;
}

#ifdef __STDC__
char	*function_winnam (char *input)
#else
char	* function_winnam(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	malloc_strcpy(&result, (curr_scr_win && curr_scr_win->name) ?
		curr_scr_win->name : empty_string);
	return result;
}

#ifdef __STDC__
char	*function_connect (char *input)
#else
char	* function_connect(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;
	char	*host;

#ifdef DAEMON_UID
	if (getuid() == DAEMON_UID)
		put_it("You are not permitted to use CONNECT()");
	else
#endif
		if ((host = new_next_arg(input, &input)) != NULL)
			result = dcc_raw_connect(host, atoi(input));
	return result;
}


#ifdef __STDC__
char	*function_listen (char *input)
#else
char	* function_listen(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

#ifdef DAEMON_UID
	if (getuid() == DAEMON_UID)
		malloc_strcpy(&result, "0");
	else
#endif
	{
		/* You know, youre just asking for trouble any time
		   you pass something to atoi() without checking it
		   first.  This is a huge coredump waiting to happen.
		 */
		if (!input && !*input)
			malloc_strcpy(&result, "0");
		else
			result = dcc_raw_listen(atoi(input));
	}
	return result;
}

#ifdef __STDC__
char	*function_toupper (char *input)
#else
char	* function_toupper(input)
	char	*input;
#endif
{
	char	*new = (char *) 0,
		*ptr;

	if (!input)
		return m_strdup(empty_string);
	malloc_strcpy(&new, input);
	for (ptr = new; *ptr; ptr++)
		*ptr = islower(*ptr) ? toupper(*ptr) : *ptr;
	return new;
}

#ifdef __STDC__
char	*function_tolower (char *input)
#else
char	* function_tolower(input)
	char	*input;
#endif
{
	char	*new = (char *) 0,
		*ptr;

	if (!input)
		return m_strdup(empty_string);
	malloc_strcpy(&new, input);
	for (ptr = new; *ptr; ptr++)
		*ptr = (isupper(*ptr)) ? tolower(*ptr) : *ptr;
	return new;
}

#ifdef __STDC__
char	*function_curpos (char *input)
#else
char	* function_curpos(input)
	char	*input;
#endif
{
	char	*new = (char *) 0,
		pos[4];

	sprintf(pos, "%d", current_screen->buffer_pos);
	malloc_strcpy(&new, pos);
	return new;
}

#ifdef __STDC__
char	*function_channels (char *input)
#else
char	* function_channels(input)
	char	*input;
#endif
{
	Window	*window;

	if (input)
		window = isdigit(*input) ? get_window_by_refnum(atoi(input))
					 : curr_scr_win;
	else
		window = curr_scr_win;

	return create_channel_list(window);
}

#ifdef __STDC__
char	*function_servers (char *input)
#else
char	* function_servers(input)
	char	*input;
#endif
{
	return create_server_list();
}

/* handled by the newer function_onchannel below */
/*
#ifdef __STDC__
char	*function_onchannel(char *input)
#else
char	* function_onchannel(input)
	char	*input;
#endif
{
	char	*nick;

	if ((char *) 0 != (nick = new_next_arg(input, &input)))
		*input++ = '\0';
	else
		return (char *) 0;
	
	if ('\0' == *input)
		return (char *) 0;

	return is_on_channel(input, nick) ? "1" : "0";
}
*/

#ifdef __STDC__
char	*function_pid(char *input)
#else
char	* function_pid(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	sprintf(buffer, "%d", (int) getpid());
	malloc_strcpy(&result, buffer);
	return result;
}

#ifdef __STDC__
char	*function_ppid (char *input)
#else
char	* function_ppid(input)
	char	*input;
#endif
{
	char	*result = (char *) 0;

	sprintf(buffer, "%d", (int) getppid());
	malloc_strcpy(&result, buffer);
	return result;
}

#ifdef __STDC__
char	*function_chanusers (char *input)
#else
char	*function_chanusers(input)
	char	*input;
#endif
{
	ChannelList	*chan;
	NickList	*nicks;
	char	*result = (char *) 0;
	int	len = 0;
	int	notfirst = 0;

	chan = lookup_channel(input, from_server, 0);
	if ((ChannelList *) 0 == chan)
		return (char *) 0;

	*buffer  = '\0';

	for (nicks = chan->nicks; nicks; nicks = nicks->next)
	{
              len += strlen(nicks->nick);
              if (len > BIG_BUFFER_SIZE)
              {
                      malloc_strcat(&result, buffer);
                      *buffer = '\0';
                      len = 0;
              }
              if (notfirst)
                      strcat(buffer, " ");
              else
                       notfirst = 1;
              strcat(buffer, nicks->nick);
	} 
	malloc_strcpy(&result, buffer);
	return result;
}

/* The new "added" functions */

/* $before(chars string of text)
 * returns the part of "string of text" that occurs before the
 * first instance of any character in "chars"
 * EX:  $before(! hop!jnelson@iastate.edu) returns "hop"
 */
#ifdef __STDC__
char	*function_before (char *word)
#else
char    *function_before(word)
char    *word;
#endif
{
        char	*booya = (char *) 0;
	char	*chars;
	char	*pointer = (char *) 0;
	int	numint;
	char 	*tmp;

	tmp = new_next_arg(word, &word);
	if (!tmp)
		return m_strdup(empty_string);

	numint = atoi(tmp);
	if (numint == 0)
	{
		numint = 1;
		chars = tmp;
	}
	else
		chars = new_next_arg(word, &word);

	if (numint < 0)
		pointer = word + strlen(word) - 1;

	pointer = search(word, &pointer, chars, numint);

	if (pointer) 
        {
		*pointer = '\0';
		malloc_strcpy(&booya, word);
	}
        else
                malloc_strcpy(&booya, empty_string);

	return booya;
}

/* $after(chars string of text)
 * returns the part of "string of text" that occurs after the 
 * first instance of any character in "chars"
 * EX: $after(! hop!jnelson@iastate.edu)  returns "jnelson@iastate.edu"
 */
#ifdef __STDC__
char	*function_after (char *word)
#else
char 	*function_after(word)
char	*word;
#endif
{
        char	*booya = (char *) 0;
	char	*chars;
	char	*pointer = (char *) 0;
	int	numint;
	char 	*tmp;

	tmp = new_next_arg(word, &word);
	if (!tmp)
		return m_strdup(empty_string);

	numint = atoi(tmp);
	if (numint)
		chars = new_next_arg(word, &word);
	else
	{
		numint = 1;
		chars = tmp;
	}

	if (numint < 0)
		pointer = word + strlen(word) - 1;

	pointer = search(word, &pointer, chars, numint);

	if (pointer) 
        {
		pointer++;
		malloc_strcpy(&booya, pointer);
	}
        else
                malloc_strcpy(&booya, empty_string);

	return booya;
}

/* $leftw(num string of text)
 * returns the left "num" words in "string of text"
 * EX: $leftw(3 now is the time for) returns "now is the"
 */
#ifdef __STDC__
char	*function_leftw (char *word)
#else
char    *function_leftw(word)
char    *word;
#endif
{
        char    *leftw_index;
        char    *booya=(char *) 0;
       	int value;
 
        leftw_index = new_next_arg(word, &word);
        if (leftw_index)
                value = atoi(leftw_index);
        else
                return m_strdup(empty_string);
        if (value < 1)
                return m_strdup(empty_string);

	booya = extract(word, 0, value-1);
        return booya;
}

/* $rightw(num string of text)
 * returns the right num words in "string of text"
 * EX: $rightw(3 now is the time for) returns "the time for"
 */
#ifdef __STDC__
char	*function_rightw (char *word)
#else
char    *function_rightw(word)
char    *word;
#endif
{
        char    *rightw_index;
        int     value;
        char    *booya=(char *)0;

        rightw_index = new_next_arg(word, &word);
        if (rightw_index)
                value = atoi(rightw_index);
        else
                return m_strdup(empty_string);
        if (value < 1)
                return m_strdup(empty_string);
                
	booya = extract(word, -value, EOS); 
        return booya;
}

/* $midw(start num string of text)
 * returns "num" words starting at word "start" in the string "string of text"
 * NOTE: The first word is word #0.
 * EX: $midw(2 2 now is the time for) returns "the time"
 */
#ifdef __STDC__
char	*function_midw (char *word)
#else
char    *function_midw(word)
char    *word;
#endif
{
        char    *midw_index;
        int     start;
        int     end;
        char    *booya= (char *)0;

        midw_index = new_next_arg(word, &word);
        if (midw_index)
                start = atoi(midw_index);
        else 
                return m_strdup(empty_string);
        midw_index = new_next_arg(word, &word);
        if (midw_index)
                end = atoi(midw_index);
        else
                return m_strdup(empty_string);
        
	booya = extract(word, start, (start + end - 1));
        return (booya);
}
        
/* $notw(num string of text)
 * returns "string of text" with word number "num" removed.
 * NOTE: The first word is numbered 0.
 * EX: $notw(3 now is the time for) returns "now is the for"
 */
#ifdef __STDC__
char	*function_notw (char *word)
#else
char    *function_notw(word)
char    *word;
#endif
{
        char    *notw_index = (char *)0;
        char    *booya = (char *)0;
        int     where;
        
        notw_index = new_next_arg(word, &word);
        if (notw_index)
                where = atoi(notw_index);
        else
                return m_strdup(empty_string);

        if (where < 0)
		return m_strdup(word);

	if (where > 0)
	{
		char *part1, *part2;
		part1 = extract(word, 0, (where - 1));
		part2 = extract(word, (where + 1), EOS);
		malloc_sprintf(&booya, "%s %s", part1, part2);
		new_free(&part1);
		new_free(&part2);
	}
	else /* where == 0 */
		booya = extract(word, 1, EOS);

        return booya;
}

/* $restw(num string of text)
 * returns "string of text" that occurs starting with and including
 * word number "num"
 * NOTE: the first word is numbered 0.
 * EX: $restw(3 now is the time for) returns "time for"
 */
#ifdef __STDC__
char	*function_restw(char *word)
#else
char    *function_restw(word)
char    *word;
#endif
{
        char    *restw_index;
        int     where;
        char    *booya=(char *)0;
        
        restw_index = new_next_arg(word, &word);
        if (restw_index)
                where = atoi(restw_index);
        else
                return m_strdup(empty_string);

	booya = extract(word, where, EOS);
        return booya;
}

/* $remw(word string of text)
 * returns "string of text" with the word "word" removed
 * EX: $remw(the now is the time for) returns "now is time for"
 */
#ifdef __STDC__
char	*function_remw (char *word)
#else
char    *function_remw(word)
char    *word;
#endif
{
        int     where;
        char    *booya = (char *) 0;
        char    *blah = (char *) 0;
        char    *lame = (char *) 0;

        malloc_strcpy(&lame, word);        
        malloc_strcpy(&blah, function_rmatch(word));

        where = atoi(blah);

	/* No matches, return the string as-is */
        if (where == 0)
        {
		char *stupid_placeholder = lame;
                new_next_arg(lame, &lame);
                new_free(&blah);
		malloc_strcpy(&stupid_placeholder, lame);
		return (stupid_placeholder);
        }
        else
        {
		char *str1, *str2;
		str1 = extract(lame, 1, (where - 1));
		str2 = extract(lame, (where + 1), EOS);
		malloc_sprintf(&booya, "%s %s", str1, str2);
		new_free(&str1);
		new_free(&str2);
                new_free(&lame);
                new_free(&blah);
        }
        return booya;
}

/* $insertw(num word string of text)
 * returns "string of text" such that "word" is the "num"th word
 * in the string.
 * NOTE: the first word is numbered 0.
 * EX: $insertw(3 foo now is the time for) returns "now is the foo time for"
 */
#ifdef __STDC__
char	*function_insertw (char *word)
#else
char    *function_insertw(word)
char    *word;
#endif
{
        char    *insertw_index;
        int     where;
        char    *what;
        char    *booya=(char *)0;
	char 	*str1, *str2;

        insertw_index = new_next_arg(word, &word);
        if (insertw_index)
                where = atoi(insertw_index);
        else
                return m_strdup(word);
        
	/* If the word goes at the front of the string, then it
	   already is: return it. ;-) */
	if (where < 1)
		malloc_strcpy(&booya, word);
	else
	{
		what = new_next_arg(word, &word);
		str1 = extract(word, 0, (where - 1));
		str2 = extract(word, where, EOS);
		malloc_sprintf(&booya, "%s %s %s",str1,what,str2);
		new_free(&str1);
		new_free(&str2);
	}

        return booya;
}

/* $chngw(num word string of text)
 * returns "string of text" such that the "num"th word is removed 
 * and replaced by "word"
 * NOTE: the first word is numbered 0
 * EX: $chngw(3 foo now is the time for) returns "now is the foo for"
 */
#ifdef __STDC__
char	*function_chngw (char *word)
#else
char    *function_chngw(word)
char    *word;
#endif
{
        char    *blah;
        int     which;
        char    *what;
        char    *booya=(char *)0;
	char	*str1, *str2;
        
        blah = new_next_arg(word,&word);
        if (blah)
                which = atoi(blah);
        else
                return m_strdup(empty_string);
                
        what = new_next_arg(word, &word);
	str1 = extract(word, 0, which - 1);
	str2 = extract(word, which + 1, EOS);
	malloc_sprintf(&booya, "%s %s %s", str1, what, str2);
	new_free(&str1);
	new_free(&str2);

        return (booya);
}

extern char    *stristr _((char *, char *));
extern char    *rstristr _((char *, char *));

/* $common (string of text / string of text)
 * Given two sets of words seperated by a forward-slash '/', returns
 * all words that are found in both sets.
 * EX: $common(one two three / buckle my two shoe one) returns "one two"
 * NOTE: returned in order found in first string.
 */
#ifdef __STDC__
char	*function_common (char *word)
#else
char    *function_common(word)
char    *word;
#endif
{
        char    *left = (char *) 0,
                *right = (char *) 0,
                *blah = (char *) 0,
                *booya = (char *) 0;

        left = word;
        if (!(right = index(word,'/')))
                return m_strdup(empty_string);

        do {
                *(right++) = '\0';
        } while (isspace(*right));

        while ((blah = new_next_arg(left, &left)))
        {
                if (stristr(right, blah))
                {
                        if (booya)
                                malloc_strcat(&booya, " ");
                        malloc_strcat(&booya, blah);
                }
        }
	if (!booya)
		malloc_strcpy(&booya, empty_string);
        return (booya);
}

/* $diff(string of text / string of text)
 * given two sets of words, seperated by a forward-slash '/', returns
 * all words that are not found in both sets
 * EX: $diff(one two three / buckle my two shoe)
 * returns "one two three buckle my shoe"
 */
#ifdef __STDC__
char	*function_diff (char *word)
#else
char    *function_diff(word)
char    *word;
#endif
{
        char    *left = (char *) 0;
        char    *right = (char *) 0;
        char    *booya = (char *) 0;
        char    *blah = (char *) 0;
        char    *leftcopy = (char *) 0;

        left = word;
        if ((right = index(word,'/')) == (char *) 0)
                return m_strdup(empty_string);

        do {
                *(right++) = '\0';
        } while (isspace(*right));

        malloc_strcpy(&leftcopy, left);

        while ((blah = new_next_arg(left, &left)))
        {
                if (!stristr(right, blah)) 
                {
                        if (booya)
                        {
                                if (!stristr(booya, blah))
                                {
                                        malloc_strcat(&booya, " ");
                                        malloc_strcat(&booya, blah);
                                }
                        }
                        else
                                malloc_strcpy(&booya, blah);
                }
      	}
	/* ok.. right now we have everything in the left side that
	 * isnt in the right side.  We now need to extract the things
	 * in the right side that arent in the left side... hmmmm
	 */
      	while ((blah = new_next_arg(right, &right)))
      	{
		/* if the word ISNT already on the list */
		if (!stristr(booya, blah))
			/* if the word ISNT on the left side */
			if (!stristr(leftcopy, blah))
			{
				/* put it in */
				malloc_strcat(&booya, " ");
				malloc_strcat(&booya, blah);
			}
      	}
	new_free(&leftcopy);
       	return (booya);
}

/* $pattern(pattern string of words)
 * given a pattern and a string of words, returns all words that
 * are matched by the pattern
 * EX: $pattern(f* one two three four five) returns "four five"
 */
#ifdef __STDC__
char	*function_pattern (char *word)
#else
char    *function_pattern(word)
char    *word;
#endif
{
        char    *blah;
        char    *booya = m_strdup(empty_string);
        char    *pattern;

        if ((pattern = new_next_arg(word, &word)))
        {
                while (((blah = new_next_arg(word, &word)) != NULL))
                {
                        if (wild_match(pattern, blah))
			{
                                if (*booya)
                                    malloc_strcat(&booya, " ");
                                malloc_strcat(&booya, blah);
			}
                }
                return (booya);
        } 
	else 
		return m_strdup(empty_string);	
}

/* $filter(pattern string of words)
 * given a pattern and a string of words, returns all words that are 
 * NOT matched by the pattern
 * $filter(f* one two three four five) returns "one two three"
 */
#ifdef __STDC__
char	*function_filter (char *word)
#else
char    *function_filter(word)
char    *word;
#endif
{
        char    *blah;
        char    *booya = m_strdup(empty_string);
        char    *pattern;

        if ((pattern = new_next_arg(word, &word)))
        {
                while ((blah = new_next_arg(word, &word)) != NULL)
                {
                        if (!wild_match(pattern, blah))
			{
                                if (*booya)
                                    malloc_strcat(&booya, " ");
                                malloc_strcat(&booya, blah);
			}
                }
                return (booya);
        } 
	else 
		return m_strdup(empty_string);
}

/* $rpattern(word list of patterns)
 * Given a word and a list of patterns, return all patterns that
 * match the word.
 * EX: $rpattern(jnelson@iastate.edu *@* jnelson@* f*@*.edu)
 * returns "*@* jnelson@*"
 */
#ifdef __STDC__
char	*function_rpattern (char *word)
#else
char    *function_rpattern(word)
char    *word;
#endif
{
        char    *blah;
        char    *booya = m_strdup(empty_string);
        char    *pattern;

        if ((blah = new_next_arg(word, &word)))
        {
                while ((pattern = new_next_arg(word, &word)) != NULL)
                {
                        if (wild_match(pattern, blah))
                        {
                                if (*booya)
                                    malloc_strcat(&booya, " ");
                                malloc_strcat(&booya, pattern);
                        }
                }
                return (booya);
        } 
	else 
		return m_strdup(empty_string);
}

/* $rfilter(word list of patterns)
 * given a word and a list of patterns, return all patterns that
 * do NOT match the word
 * EX: $rfilter(jnelson@iastate.edu *@* jnelson@* f*@*.edu)
 * returns "f*@*.edu"
 */
#ifdef __STDC__
char	*function_rfilter (char *word)
#else
char    *function_rfilter(word)
char    *word;
#endif
{
        char    *blah;
        char    *booya = m_strdup(empty_string);
        char    *pattern;

        if ((blah = new_next_arg(word, &word)))
        {
                while ((pattern = new_next_arg(word, &word)) != NULL)
                {
                        if (!wild_match(pattern, blah))
                        {
                                if (*booya)
                                    malloc_strcat(&booya, " ");
                                malloc_strcat(&booya, pattern);
                         }
                }
                return (booya);
        } 
	else 
		return m_strdup(empty_string);
}

/* $copattern(pattern var_1 var_2)
 * Given a pattern and two variable names, it returns all words
 * in the variable_2 corresponding to any words in variable_1 that
 * are matched by the pattern
 * EX: @nicks = [hop IRSMan skip]
 *     @userh = [jnelson@iastate.edu irsman@iastate.edu sanders@rush.cc.edu]
 *     $copattern(*@iastate.edu nicks userh) 
 *	returns "jnelson@iastate.edu irsman@iastate.edu"
 */
#ifdef __STDC__
char	*function_copattern (char *word)
#else
char    *function_copattern(word)
char    *word;
#endif
{
	char	*booya = (char *) 0,
		*pattern = (char *) 0,
		*firstl = (char *) 0, *firstlist = (char *) 0, *firstel = (char *) 0,
		*secondl = (char *) 0, *secondlist = (char *) 0, *secondel = (char *) 0;
	char	*flp, *slp;

	if ((pattern = new_next_arg(word, &word)))
	{
		if ((firstlist = new_next_arg(word, &word)))
		{
			if ((secondlist = new_next_arg(word, &word)))
			{
				firstl=find_inline(firstlist);
				secondl=find_inline(secondlist);
				flp = firstl;
				slp = secondl;
			} else 
				return m_strdup(empty_string);
		} else 
			return m_strdup(empty_string);
	} else 
		return m_strdup(empty_string);
	
	while ((firstel = new_next_arg(firstl, &firstl)))
	{
		if (!(secondel = new_next_arg(secondl, &secondl)))
		{
			new_free(&flp);
			new_free(&slp);
			return (booya);
		}
                if (wild_match(pattern, firstel))
                {
                        if (booya)
                           malloc_strcat(&booya, " ");
                        malloc_strcat(&booya, secondel);
                }
        }
	new_free(&flp);
	new_free(&slp);
        if (!booya) 
		malloc_strcpy(&booya, empty_string);

        return (booya);
}

/* $beforew(pattern string of words)
 * returns the portion of "string of words" that occurs before the 
 * first word that is matched by "pattern"
 * EX: $beforew(three one two three o leary) returns "one two"
 */
#ifdef __STDC__
char	*function_beforew (char *word)
#else
char    *function_beforew(word)
char    *word;
#endif
{
        int     where;
        char    *booya = (char *) 0;
	char	*lame = (char *) 0;

	malloc_strcpy(&lame, word);
        where = atoi(function_rmatch(word));
	if (where < 1)
		malloc_strcpy(&booya, empty_string);
	else
		booya = extract(lame, 1, where - 1);
	new_free(&lame);
        return booya;
}
                
/* Same as above, but includes the word being matched */
#ifdef __STDC__
char	*function_tow (char *word)
#else
char    *function_tow(word)
char    *word;
#endif
{
        int     where;
        char    *booya = (char *) 0;
	char	*lame = (char *) 0;
        
	malloc_strcpy(&lame, word);
        where = atoi(function_rmatch(word));
	if (where < 1)
		malloc_strcpy(&booya, empty_string);
	else
		booya = extract(lame, 1, where);
	new_free(&lame);
        return booya;
}

/* Returns the string after the word being matched */
#ifdef __STDC__
char	*function_afterw (char *word)
#else
char    *function_afterw(word)
char    *word;
#endif
{
        int     where;
        char    *booya = (char *) 0;
	char	*lame = (char *) 0;
        
	malloc_strcpy(&lame, word);
        where = atoi(function_rmatch(word));
	if (where < 1)
		malloc_strcpy(&booya, empty_string);
	else
		booya = extract(lame, where, EOS);
	new_free(&lame);
        return booya;
}

/* Returns the string starting with the word being matched */
#ifdef __STDC__
char	*function_fromw (char *word)
#else
char    *function_fromw(word)
char    *word;
#endif
{
        int     where;
        char    *booya = (char *) 0;
	char 	*lame = (char *) 0;

	malloc_strcpy(&lame, word);
        where = atoi(function_rmatch(word));
	if (where < 1)
		malloc_strcpy(&booya, empty_string);
	else
		booya = extract(lame, where+1, EOS);
	new_free(&lame);
        return booya;
}

/* Cut and paste a string */
#ifdef __STDC__
char	*function_splice (char *word)
#else
char    *function_splice(word)
char    *word;
#endif
{
        char    *array = (char *) 0;
        int	start;
	int	length;
	char	*replacement = (char *) 0;
	char	*booya = (char *) 0;
	char	*blah = (char *) 0;
	char	*goofy = (char *) 0;
	char	*str1;
	int 	old_display = window_display;

	array = new_next_arg(word, &word);
	goofy = find_inline(array);
        if (!goofy || !*goofy)
                return m_strdup(empty_string);

	if (word) 
                start = atoi(new_next_arg(word, &word));
	else 
        {
                new_free(&goofy);
                return m_strdup(empty_string);
        }

	if (word)
		length = atoi(new_next_arg(word, &word));
	else
	{
		new_free(&goofy);
		return m_strdup(empty_string);
	}

	if (word) 
                replacement = word;
        if (*replacement)
                malloc_strcat(&replacement, " ");

	booya = extract(goofy, start, start+length-1);

        if (start)
        {
		blah = extract(goofy, 0, start-1);
	        malloc_strcat(&blah, " ");
        }
        else
                new_free(&blah);

	malloc_strcat(&blah, replacement);
	str1 = extract(goofy, start+length, EOS);
	malloc_strcat(&blah, str1);
	new_free(&str1);

	window_display = 0;
	add_alias(VAR_ALIAS, array, blah);
	window_display = old_display;

	new_free(&blah);
	new_free(&goofy);
	new_free(&replacement);
	return (booya);
}
	
#ifdef __STDC__
char	*function_numonchannel (char * word)
#else
char    *function_numonchannel(word)
char    *word;
#endif
{
        char            *channel = (char *) 0;
        ChannelList	*chan = (ChannelList *) 0;
	NickList	*tmp = (NickList *) 0;
	char		*booya = (char *) 0;
        int             counter = 0;
	
	channel = new_next_arg(word, &word);
	if ((channel == (char *) 0) || (!is_channel(channel)) ||
	    (!(chan = lookup_channel(channel, curr_scr_win->server,0))))
		return m_strdup("0");

        for (tmp = chan->nicks; tmp; tmp = tmp->next)
		counter++;

	malloc_sprintf(&booya, "%d", counter);
	return booya;
}
	
#ifdef __STDC__
char	*function_onchannel (char *word)
#else
char    *function_onchannel(word)
char    *word;
#endif
{
        char            *channel = (char *) 0;
        ChannelList     *chan = (ChannelList *) 0;
        NickList        *tmp = (NickList *) 0;
        char            *nicks = (char *) 0;
	int		free = 0;

        channel = new_next_arg(word, &word);
        if (channel == (char *) 0)
	{
		free = 1;
                malloc_strcpy(&channel, get_channel_by_refnum(0));
	}
	else
        if (!strcmp(channel, "*"))
	{
		free = 1;
                malloc_strcpy(&channel, get_channel_by_refnum(0));
	}

        if ((channel == (char *) 0) || (*channel == '0'))
	{
		if (free) new_free(&channel);
                return m_strdup(empty_string);
	}

        if (!is_channel(channel))   /* must be a new onchannel */
        {
                char *nick = channel;
                if ((channel = new_next_arg(word, &word)))
	                return m_strdup(is_on_channel(channel, nick) ? "1" : "0");
	        else
		        return m_strdup("0");
        }

        chan = lookup_channel(channel,curr_scr_win->server,0);
        if (chan == (ChannelList *) 0)
                return m_strdup(empty_string);

        for (tmp = chan->nicks; tmp; tmp = tmp->next)
        {
                if (nicks)
                    malloc_strcat(&nicks, " ");
                malloc_strcat(&nicks, tmp->nick);
        }
	if (free) new_free(&channel);
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
        return(nicks);
}

#ifdef __STDC__
char    *function_chops(char *word)
#else
char    *function_chops(word)
char    *word;
#endif
{
        char            *channel = (char *) 0;
        ChannelList     *chan;
        NickList        *tmp;
        char            *nicks = (char *) 0;
	int 		free = 0;

        if ((channel = new_next_arg(word, &word)) == (char *) 0)
	{
		free = 1;
                malloc_strcpy(&channel, get_channel_by_refnum(0));
	}
        else
        if (!strcmp(channel, "*"))
	{
		free = 1;
                malloc_strcpy(&channel, get_channel_by_refnum(0));
	}

        if ((channel == (char *) 0) || (*channel == '0'))
	{
		if (free) new_free(&channel);
                return m_strdup(empty_string);
	}

        chan = lookup_channel(channel,curr_scr_win->server,0);
        if (!chan)
	{
		if (free) new_free(&channel);
		return m_strdup(empty_string);
	}

        for (tmp = chan->nicks; tmp; tmp = tmp->next)
        {
                if (tmp->chanop == 1) 
                {
                        if (nicks)
                            malloc_strcat(&nicks, " ");
                        malloc_strcat(&nicks, tmp->nick);
                }
        }
	if (free) new_free(&channel);
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
        return(nicks);
}        

#ifdef __STDC__
char    *function_nochops(char *word)
#else
char    *function_nochops(word)
char    *word;
#endif
{
        char            *channel = (char *)0;
        ChannelList     *chan;
        NickList        *tmp;
        char            *nicks = (char *)0;
	int		free = 0;

        if ((channel = new_next_arg(word, &word)) == (char *) 0)
	{
		free = 1;
                malloc_strcpy(&channel,get_channel_by_refnum(0));
	}

        if (!strcmp(channel, "*"))
	{
		free = 1;
                malloc_strcpy(&channel,get_channel_by_refnum(0));
	}

        chan = lookup_channel(channel,curr_scr_win->server,0);
        if (!chan)
	{
		if (free) new_free(&channel);
		return m_strdup(empty_string);
	}
        for (tmp = chan->nicks; tmp; tmp = tmp->next)
        {
                if (tmp->chanop == 0) {
                        if (nicks)
                            malloc_strcat(&nicks, " ");
                        malloc_strcat(&nicks, tmp->nick);
                }
        }
	if (free) new_free(&channel);
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
        return(nicks);
}

#ifdef __STDC__
char    *function_key(char *word)
#else
char    *function_key(word)
char    *word;
#endif
{
        char    *channel = (char *) 0;
        char    *booya = (char *) 0;

        channel = new_next_arg(word, &word);
        do
        {
                char    *key;

                if (!channel)
                        if (booya)
                                break;
                        else
                        malloc_strcpy(&channel, get_channel_by_refnum(0));

                if (!strcmp(channel, "*"))
                        malloc_strcpy(&channel, get_channel_by_refnum(0));

                if ((key = channel_key(channel)))
                {
                        if (booya)
                            malloc_strcat(&booya, " ");
                        malloc_strcat(&booya, key);
                }
        }
        while ((channel = new_next_arg(word, &word)));

	new_free(&channel);
	if (!booya)
		malloc_strcpy(&booya, empty_string);
        return (booya);
}

#ifdef __STDC__
char    *function_revw (char *words)
#else
char    *function_revw (words)
char    *words;
#endif
{
	char	**thewords;
        char    *booya = (char *) 0;
        int     x = 0;

	thewords = (char **)new_malloc(word_count(words) * sizeof(char *) + 1);
        while ((thewords[x++] = new_next_arg(words, &words)))
                ;

        for (x--; x >= 0; x--)
        {
                if (booya)
                    malloc_strcat(&booya," ");
                malloc_strcat(&booya,thewords[x]);
        }
	if (!booya)
		malloc_strcpy(&booya, empty_string);
        return (booya);
}

#ifdef __STDC__
char    *function_reverse(char *words)
#else
char    *function_reverse(words)
char    *words;
#endif
{
        int     length = strlen(words);
        char    *booya = (char *) 0;
        int     x = 0;

        booya = (char *)new_malloc(length+1);
        for(length--; length >= 0; length--,x++)
                booya[x] = words[length];
        booya[x]='\0';
        return (booya);
}

#ifdef __STDC__
char    *function_jot(char *word)
#else
char    *function_jot(word)
char    *word;
#endif
{
        int     start = 0;
        int     stop = 0;
        int     interval = 1;
        int     counter;
        char    booya[4096];
        char    *foobar = (char *) 0;
        char    *foob = (char *) 0;

	bzero(booya, 4096);

        if ((foobar = new_next_arg(word, &word)))
        {
                start = atoi(foobar);
                if ((foobar = new_next_arg(word, &word)))
                {
                        stop = atoi(foobar);
                        if ((foobar = new_next_arg(word, &word)))
                                interval = atoi(foobar);
                        else
                                interval = 1;
                }
                else
                        return m_strdup(empty_string);
        }
        else
                return m_strdup(empty_string);

        if (interval < 0) 
                interval = -interval;
        if (start < stop)
                for (counter = start;counter <= stop;counter+=interval)
                {
                        if (booya && *booya)
	                        sprintf(booya, "%s %d",booya,counter);
			else
				sprintf(booya, "%d", counter);
		}
        else
                for (counter = start;counter >= stop;counter-=interval)
		{
			if (booya && *booya)
	                        sprintf(booya, "%s %d",booya,counter);
			else
				sprintf(booya, "%d", counter);
		}
	booya[4095] = '\0';
        malloc_strcpy(&foob, booya);
        return (foob);
}

#ifdef __STDC__
char    *function_shift(char *word)
#else
char    *function_shift(word)
char	*word;
#endif
{
        char    *value = (char *) 0;
        char    *var    = (char *) 0;
	char	*booya 	= (char *) 0;
	char	*blah;
        int     old_display = window_display;
        char    *placeholder;

	var = new_next_arg(word, &word);
        if (!var)
                return m_strdup(empty_string);
	if (word && *word)
	{
		malloc_strcpy(&value, var);
		malloc_strcat(&value, word);
		var = (char *) 0;
	}
	else
		value = find_inline(var);
        if (!value && !*value)
                return m_strdup(empty_string);
        placeholder = value;
	blah = new_next_arg(value, &value);
	malloc_strcpy(&booya, blah);
	if (var)
	{
		window_display = 0;
		add_alias(VAR_ALIAS, var, value);
		window_display = old_display;
	}
        new_free(&placeholder);
	return booya;
}

#ifdef __STDC__
char    *function_unshift(char *word)
#else
char    *function_unshift(word)
char	*word;
#endif
{
        char    *value = (char *) 0;
        char    *var    = (char *) 0;
	char	*booya  = (char *) 0;
        int     old_display = window_display;

	var = new_next_arg(word, &word);
        if (!var)
                return m_strdup(empty_string);
	value = find_inline(var);
        if (!word)
                return value;
	malloc_sprintf(&booya, "%s %s", word, value);
        window_display = 0;
	add_alias(VAR_ALIAS, var, booya);
        window_display = old_display;
        new_free(&value);
	return (booya);
}

#ifdef __STDC__
char    *function_push(char *word)
#else
char    *function_push(word)
char	*word;
#endif
{
        char    *value = (char *) 0;
        char    *var    = (char *) 0;
        int     old_display = window_display;

	var = new_next_arg(word, &word);
        if (!var)
                return m_strdup(empty_string);
	value = find_inline(var);
        if (value && *value)
	        malloc_strcat(&value, " ");
	malloc_strcat(&value, word);
        window_display = 0;
	add_alias(VAR_ALIAS, var, value);
        window_display = old_display;
	return (value);
}

#ifdef __STDC__
char	*function_pop(char *word)
#else
char	*function_pop(word)
char	*word;
#endif
{
	char *value	= (char *) 0;
	char *var	= (char *) 0;
	char *pointer	= (char *) 0;
        int     old_display = window_display;
        char *blech     = (char *) 0;

	var = new_next_arg(word, &word);
        if (!var)
                return m_strdup(empty_string);
	if (word && *word)
	{
		malloc_strcpy(&value, var);
		malloc_strcpy(&value, word);
		var = (char *) 0;
	}
	else
		value = find_inline(var);

	if (!value || !*value)
	{
		new_free(&value);
		return m_strdup(empty_string);
	}
	pointer = rindex(value, ' ');
	if (!pointer)
	{
		window_display = 0;
		add_alias(VAR_ALIAS, var, empty_string); /* dont forget this! */
		window_display = 1;
		return value;	/* one word -- return it */
	}
	*pointer++ = '\0';
        window_display = 0;
	add_alias(VAR_ALIAS, var, value);
        window_display = old_display;
        malloc_strcpy(&blech, pointer);
        new_free(&value);
	return (blech);
}

/* Search and replace function --
   Usage:   $sar(c/search/replace/data)
   Commands:
   		r - treat data as a variable name and 
		    return the replaced data to the variable
                g - Replace all instances, not just the first one
   The delimiter may be any character that is not a command (typically /)
   The delimiter MUST be the first character after the command
   Returns empty string on error
*/

#ifdef __STDC__
char    *function_sar(char *word)
#else
char    *function_sar(word)
char    *word;
#endif
{
        char    delimiter;
	char	*pointer	= (char *) 0;
        char    *search         = (char *) 0;
        char    *replace        = (char *) 0;
        char    *data		= (char *) 0;
	char	*value		= (char *) 0;
	char	*booya		= (char *) 0;
        char    *placeholder    = (char *) 0;
	int	variable = 0,global = 0,searchlen,oldwindow;

	while ((*word == 'r') || (*word == 'g'))
	{
		if (*word == 'r') {
			variable = 1;
			*word++ = '\0';
		}
		if (*word == 'g') {
			global = 1;
			*word++ = '\0';
		}
	}
	delimiter = *word++;
	search = word;
	if ((replace = index(word,delimiter)) == 0)
		return m_strdup(empty_string);

	*replace++ = '\0';
	if ((data = index(replace,delimiter)) == 0)
		return m_strdup(empty_string);

	*data++ = '\0';

	if (variable == 1)
        {
		value = find_inline(data);
                placeholder = value;
        }
	else
		value = data;

	if ((!value) || (!*value))
		return m_strdup(empty_string);

	pointer = value;
	searchlen = strlen(search) - 1;

        if (global)
	{
		while ((pointer = strstr(pointer,search)) != (char *) 0)
		{
			*pointer = '\0';
			pointer += searchlen;
			*pointer++ = '\0';
			malloc_strcat(&booya, value);
			malloc_strcat(&booya, replace);
			value = pointer;
		}
	} else
	{
		if ((pointer = strstr(pointer,search)) != (char *) 0)
		{
			*pointer = '\0';
			pointer += searchlen;
			*pointer++ = '\0';
			malloc_strcat(&booya, value);
			malloc_strcat(&booya, replace);
			value = pointer;
		}
	}
	malloc_strcat(&booya, value);
        if (variable == 1) 
        {
                oldwindow = window_display;
                window_display = 0;
                add_alias(VAR_ALIAS, data, booya);
                window_display = oldwindow;
        }
        new_free(&placeholder);
	return (booya);
}

#ifdef __STDC__
char    *function_center(char *word)
#else
char    *function_center(word)
char    *word;
#endif
{
	char	*width = (char *) 0;
	int	length,pad,x=0,widthc;
	char 	*padc;
	char	*booya = (char *) 0;

	width = new_next_arg(word, &word);
        if (width && *width)
                widthc = atoi(width);
        else return m_strdup(empty_string);

	length = strlen(word);
	
	if ((pad = widthc - length) < 0)
		return (word);

	pad /= 2;
	padc = (char *)new_malloc(pad+1);
	
	while (x<pad)
		padc[x++] = ' ';
        *(padc+pad) = '\0';

	malloc_strcpy(&booya, padc);
	malloc_strcat(&booya, word);
	new_free(&padc);
	return (booya);
}

#ifdef __STDC__
char    *function_split(char *word)
#else
char    *function_split(word)
char    *word;
#endif
{
	char	*chrs;
	char	*pointer;
	char	*booya = (char *) 0;

	chrs = new_next_arg(word, &word);
	while ( (pointer = sindex(word,chrs)) != (char *) 0 )
		*pointer++ = ' ';
	malloc_strcpy(&booya, word);
	return booya;
}

#ifdef __STDC__
char *function_chr(char *word)
#else
char *function_chr(word)
char *word;
#endif
{
        char *booya = (char *) 0;
        char aboo[1024];
        char *ack = aboo;
        char *blah;

        while ((blah = next_arg(word, &word)))
        {
                int value;

                if (blah)
                        value = atoi(blah);

                *ack++ = (char) value;
        }
        *ack = '\0';
        malloc_strcpy(&booya, aboo);
        return booya;
}

#ifdef __STDC__
char *function_ascii(char *word)
#else
char *function_ascii(word)
char *word;
#endif
{
	char *aboo = (char *) 0;
        char *ack = word;

        do
        {
                int ick = (int) *ack++;
			/* Dont fret, i purposedly wrote malloc_sprintf
			 * to make this line here safe...
			 */
                if (aboo)
			malloc_sprintf(&aboo, "%s %d",aboo, ick);
                else
			malloc_sprintf(&aboo, "%d",ick);
        }
        while (*ack);

        return aboo;
}

#ifdef __STDC__
char	*function_which(char *word)
#else
char	*function_which(word)
char	*word;
#endif
{
	char *file1 = (char *) 0;
	char *irc_path;
	FILE *fp;

	irc_path = get_string_var(LOAD_PATH_VAR);

	if (word)
	{
		malloc_strcpy(&file1, new_next_arg(word, &word));
		if (word && *word)
			malloc_strcpy(&irc_path, word);

		fp = uzfopen (&file1, irc_path);
		if (fp)
		{
			fclose (fp);
			return (file1);
		}
		else
		{
			new_free(&file1);
			return m_strdup (empty_string);
		}
	}
	else
		return m_strdup(empty_string);
}


#ifdef __STDC__
char *function_isalpha(char *words)
#else
char *function_isalpha(words)
char *words;
#endif
{
	char *booya = (char *) 0;

	if (((*words >= 'a') && (*words <= 'z')) ||
	    ((*words >= 'A') && (*words <= 'Z')))
		malloc_strcpy(&booya, "1");
	else
		malloc_strcpy(&booya, "0");
	return booya;
}

#ifdef __STDC__
char *function_isdigit(char *words)
#else
char *function_isdigit(words)
char *words;
#endif
{
	char *booya = (char *) 0;
	if (((*words >= '0') && (*words <= '9')) ||
	    ((*words == '-') && ((*(words+1) >= '0') || (*(words+1) <= '9'))))
		malloc_strcpy(&booya, "1");
	else
		malloc_strcpy(&booya, "0");
	return booya;
}

/* stuff from files.c.... */
int open_file_for_read _((char *));
int open_file_for_write _((char *));
int file_write _((int, char *));
char *file_read _((int));
int file_eof _((int));
int file_close _((int));

#ifdef __STDC__
char *function_open (char *words)
#else
char *function_open (words)
char *words;
#endif
{
	char *filename = new_next_arg(words, &words);
	char *type = next_arg(words, &words);
	char *blah = (char *) 0;

	if (type && *type && toupper(*type) == 'R')
		malloc_sprintf(&blah, "%d", open_file_for_read(filename));
	else if (type && *type && toupper(*type) == 'W')
		malloc_sprintf(&blah, "%d", open_file_for_write(filename));

	if (!blah)
		malloc_strcpy(&blah, empty_string);

	return blah;
}

#ifdef __STDC__
char *function_close (char *words)
#else
char *function_close (words)
char *words;
#endif
{
	char *fdc = new_next_arg(words, &words);
	int fd;
	char *blah = (char *) 0;

	if (fdc && *fdc)
		fd = atoi(fdc);
	else
		return m_strdup(empty_string);

	malloc_sprintf(&blah, "%d", file_close(fd));

	return blah;
}	

#ifdef __STDC__
char *function_write (char *words)
#else
char *function_write (words)
char *words;
#endif
{
	char *fdc = new_next_arg(words, &words);
	int fd;
	char *blah = (char *) 0;

	if (fdc && *fdc)
		fd = atoi(fdc);
	else
		return m_strdup(empty_string);

	malloc_sprintf(&blah, "%d", file_write(fd, words));
	return blah;
}

#ifdef __STDC__
char *function_read (char *words)
#else
char *function_read (words)
char *words;
#endif
{
	char *fdc = new_next_arg(words, &words);
	int fd;

	if (fdc && *fdc)
		fd = atoi(fdc);
	else
		return m_strdup(empty_string);

	return file_read (fd);
}

#ifdef __STDC__
char *function_eof (char *words)
#else
char *function_eof (words)
char *words;
#endif
{
	char *fdc = new_next_arg(words, &words);
	int fd;
	char *blah = (char *) 0;

	if (fdc && *fdc)
		fd = atoi(fdc);
	else
		return m_strdup(empty_string);

	malloc_sprintf(&blah, "%d", file_eof(fd));
	return blah;
}

#ifdef __STDC__
char *host_to_ip (const char *host)
#else
char *host_to_ip (host)
const char *host;
#endif
{
	static char ip[256];
	struct hostent *hep;

	alarm(1);
	hep = gethostbyname(host);
	alarm(0);

	return (hep ? 
		sprintf(ip,"%u.%u.%u.%u", 	hep->h_addr[0] & 0xff,
						hep->h_addr[1] & 0xff,
						hep->h_addr[2] & 0xff,
						hep->h_addr[3] & 0xff),
		ip : empty_string);
}

#ifdef __STDC__
char *ip_to_host (const char *ip)
#else
char *ip_to_host (ip)
const char *ip;
#endif
{
	int b1 = 0, b2 = 0, b3 = 0, b4 = 0;
	char foo[4];
	struct hostent *hep;
	static char host[257];

	sscanf(ip,"%d.%d.%d.%d", &b1, &b2, &b3, &b4);
	foo[0] = b1;
	foo[1] = b2;
	foo[2] = b3;
	foo[3] = b4;

	alarm(1);
	hep = gethostbyaddr(foo, 4, AF_INET);
	alarm(0);

	return (hep ?
		strcpy(host, hep->h_name): 
		empty_string);
}

#ifdef __STDC__
char *one_to_another (const char *what)
#else
char *one_to_another (what)
const char *what;
#endif
{
	int b1 = 0, b2 = 0, b3 = 0, b4 = 0;
	sscanf(what, "%d.%d.%d.%d", &b1, &b2, &b3, &b4);

	if ((b1 == 0) && (b2 == 0) && (b3 == 0) && (b4 == 0))
		return host_to_ip (what);
	else
		return ip_to_host (what);
}


#ifdef __STDC__
char *function_iptoname (char *words)
#else
char *function_iptoname (words)
char *words;
#endif
{
	return m_strdup(ip_to_host(words));
}

#ifdef __STDC__
char *function_nametoip (char *words)
#else
char *function_nametoip (words)
char *words;
#endif
{
	return m_strdup(host_to_ip(words));
}

#ifdef __STDC__
char *function_convert (char *words)
#else
char *function_convert (words)
char *words;
#endif
{
	return m_strdup(one_to_another(words));
}

#ifdef __STDC__
char *function_translate (char *words)
#else
char *function_translate (words)
char *words;
#endif
{
	char *oldc, *newc, *text;
	char *ptr;
	int size_old, size_new;
	int x;
	char delim;

	oldc = words;
	/* First character can be a slash.  If it is, we just skip over it */
	delim = *oldc;
	oldc++;
	newc = strchr(oldc, delim);
	if (newc)
	{
		newc++;
		text = strchr(newc, delim);
		if (text)
		{
			/* Dont forget to null out the delim */
			*text = '\0';
			text++;
		}
		else
			/* No text in, no text out */
			return m_strdup(empty_string);

		/* Right here we do a sanity check to make sure that
		   oldc isnt pointing at a delim.  If it is, that means
		   there are no characters to replace.  So we punt */
		if (*oldc == delim)
			return m_strdup(text);

		/* If newc points at a '/', then that means that the
		   characters to replace with is empty -- ie, the oldc
		   characters are to be removed (ala $strip()).  It wouldnt
		   make any sense to have it pointing at '/', so we make
		   it point to empty_string, which is more appropriate. */
		if (*newc == delim)
		{
			/* Dont forget to null out the delim */
			*newc = '\0';
			newc = empty_string;
		}
		else
			newc[-1] = 0;

		/* this is cheating, but oh well, >;-) */
		text = m_strdup(text);

		size_new = strlen(newc);
		size_old = strlen(oldc);

		for (ptr = text; ptr && *ptr; ptr++)
		{
			for (x = 0;x < size_old;(void)x++)
			{
				if (*ptr == oldc[x])
				{
					/* Check to make sure we arent
					   just eliminating the character.
					   If we arent, put in the new char,
					   otherwise strcpy it away */
					if (size_new)
						*ptr = newc[(x<size_new)?x:size_new-1];
					else
						strcpy (ptr, ptr+1);
					break;
				}
			}
		}
		return text;
	}
	else
		/* No text in, no text out */
		return m_strdup(empty_string);
}

#ifdef __STDC__
char *function_server_version (char *word)
#else
char *function_server_version (word)
char *word
#endif
{
	int servnum;
	int version;

	if (word && *word)
		servnum = atoi(next_arg(word, &word));
	else
		servnum = primary_server;

	if (servnum > number_of_servers)
		return m_strdup("unknown");

	version = get_server_version(servnum);

#ifdef COMPAT_27
	if (version == Server2_7)		return m_strdup("2.7");
	else
#endif
	if (version == Server2_8) 		return m_strdup("2.8");
	else if (version == Server_u2_8) 	return m_strdup("u2.8");
	else if (version == Server_u2_9) 	return m_strdup("u2.9");
	else if (version == Server2_9) 		return m_strdup("2.9");
	else if (version == Server_u3_0)	return m_strdup("u3.0");

	return m_strdup("Unknown");
}

#ifdef __STDC__
char *function_unlink(char *words)
#else
char *function_unlink (words)
char *words
#endif
{
	char *filename;
	char booya[10];
	int 	failure = 0;

	while ((filename = next_arg(words, &words)))
		failure -= unlink(filename);	
	sprintf(booya, "%d", failure);
	return m_strdup(booya);
}

#ifdef __STDC__
char *function_rename (char *words)
#else
char *function_rename (words)
char *words
#endif
{
	char *filename1, *filename2;
	char booya[10];
	int 	failure = 0;

	filename1 = next_arg(words, &words);
	filename2 = next_arg(words, &words);
	failure = filename2 ? rename(filename1, filename2) : 0;
	sprintf(booya, "%d", failure);
	return m_strdup(booya);
}

#ifdef __STDC__
char *function_rmdir (char *words)
#else
char *function_rmdir (words)
char *words
#endif
{
	char *filename;
	char booya[10];
	int 	failure = 0;

	while ((filename = next_arg(words, &words)))
		failure -= rmdir(filename);	
	sprintf(booya, "%d", failure);
	return m_strdup(booya);
}

#ifdef __STDC__
char *function_truncate (char *words)
#else
char *function_truncate (words)
char *words
#endif
{
	int num = 0;
	char *num_c;
	double value = 0;
	char *value_c;
	char buffer[1024];
	char format[1024];

	num_c = next_arg(words, &words);
	if (num_c && *num_c)
		num = atol(num_c);

	value_c = next_arg(words, &words);
	if (value_c && *value_c)
		value = atof(value_c);

	if (num < 0)
	{
		float foo;
		int end;

		sprintf(format, "%%.%de", -num-1);
		sprintf(buffer, format, value);
		foo = atof(buffer);
		sprintf(buffer, "%lf", foo);
		end = strlen(buffer) - 1;
		if (end == 0)
			return m_strdup(empty_string);
		while (buffer[end] == '0')
			end--;
		if (buffer[end] == '.')
			end--;
		buffer[end+1] = 0;
	}
	else if (num > 0)
	{
		sprintf(format, "%%10.%dlf", num);
		sprintf(buffer, format, value);
	}
	else
		return m_strdup(empty_string);

	while (*buffer == ' ')
		strcpy(buffer, buffer+1);

	return m_strdup(buffer);
}
