/*
 * ircII: a new irc client.  I like it.  I hope you will too!
 *
 * Written By Michael Sandrof
 * Copyright(c) 1990 
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 *
 * Reformatted by Matthew Green, November 1992.
 */
#define IRCII_VERSION	"2.8.2-EPIC2.003"

/*
 * INTERNAL_VERSION is the number that the special alias $V returns.
 * Make sure you are prepared for floods, pestilence, hordes of locusts, 
 * and all sorts of HELL to break loose if you change this number.
 * Its format is actually YYYYMMDD, for the _release_ date of the
 * client..
 */
#define INTERNAL_VERSION	"19950601"


#include "irc.h"
 
#include <sys/stat.h>
#include <pwd.h>
 
#ifdef ISC22
# include <sys/bsdtypes.h>
#endif /* ISC22 */
 
#ifdef ESIX
# include <lan/net_types.h>
#endif /* ESIX */
 
#if USING_CURSES
# include <curses.h>
#endif /* POSIX */
 
#if DO_USER2
# include <setjmp.h>
#endif

#include "irc.h"
#include "status.h"
#include "dcc.h"
#include "names.h"
#include "vars.h"
#include "input.h"
#include "alias.h"
#include "output.h"
#include "term.h"
#include "exec.h"
#include "screen.h"
#include "log.h"
#include "server.h"
#include "hook.h"
#include "keys.h"
#include "ircaux.h"
#include "edit.h"
#include "window.h"
#include "history.h"
#include "exec.h"
#include "notify.h"
#include "mail.h"
#include "debug.h"

int	irc_port = IRC_PORT,			/* port of ircd */
	send_text_flag = -1,			/* used in the send_text()
						 * routine */
	use_flow_control = USE_FLOW_CONTROL,	/* true: ^Q/^S used for flow
						 * cntl */
	irc_io_loop = 1,			/* see irc_io below */
	break_io_processing = 0,		/* more forceful than
						 * irc_io_loop */
	current_numeric,			/* this is negative of the
						 * current numeric! */
	dumb = 0,				/* if true, IRCII is put in
						 * "dumb" mode */
	no_fork = 0,				/* if true, IRCII won't with
						 * -b or -e */
        background = 0,                         /* if true, ircII will fork */
	use_input = 1,				/* if 0, stdin is never
						 * checked */
	waiting = 0,				/* used by /WAIT command */
	who_mask = 0;				/* keeps track of which /who
						 * switchs are set */

char	oper_command = 0;	/* true just after an oper() command is
				 * given.  Used to tell the difference
				 * between an incorrect password generated by
				 * an oper() command and one generated when
				 * connecting to a new server */
char	global_all_off[2];		/* lame kludge to get around lameness */
char	MyHostName[80];			/* The local machine name. Used by
					 * DCC TALK */
	struct	in_addr	MyHostAddr;		/* The local machine address */
	struct	in_addr	local_ip_address; /* Sometimes the same, sometimes not */
extern	char	*last_away_nick;

char	*invite_channel = (char *) 0,		/* last channel of an INVITE */
	buffer[BIG_BUFFER_SIZE + 1],		/* multipurpose buffer */
	*ircrc_file = (char *) 0,		/* full path .ircrc file */
	*my_path = (char *) 0,		/* path to users home dir */
	*irc_path = (char *) 0,		/* paths used by /load */
	*irc_lib = (char *) 0,		/* path to the ircII library */
	nickname[NICKNAME_LEN + 1],		/* users nickname */
	hostname[NAME_LEN + 1],			/* name of current host */
	realname[REALNAME_LEN + 1],		/* real name of user */
	username[NAME_LEN + 1],			/* usernameof user */
	*send_umode = NULL,			/* sent umode */
	*args_str = (char *) 0,		/* list of command line args */
	*last_notify_nick = (char *) 0,	/* last detected nickname */
	empty_string[] = "",			/* just an empty string */
	*who_name = (char *) 0,		/* extra /who switch info */
	*who_file = (char *) 0,		/* extra /who switch info */
	*who_server = (char *) 0,		/* extra /who switch info */
	*who_host = (char *) 0,		/* extra /who switch info */
	*who_nick = (char *) 0,		/* extra /who switch info */
	*who_real = (char *) 0,		/* extra /who switch info */
	*cannot_open = (char *) 0,		/* extra /who switch info */
	*cut_buffer = (char *) 0;		/* global cut_buffer */

	int away_set = 0;			/* set if there is an away
						 * message anywhere */
	int	quick_startup = 0;		/* set if we ignore .ircrc */
	time_t	idle_time = 0;
        time_t  start_time;

	RETSIGTYPE	cntl_c();
	RETSIGTYPE	sig_user1() ;
#ifdef CORECATCH
	RETSIGTYPE	coredump() ;
#endif /* CORECATCH */

static	int	cntl_c_hit = 0;

#if DO_USER2
	jmp_buf	outta_here;
#endif

	char	irc_version[] = IRCII_VERSION;
	char	internal_version[] = INTERNAL_VERSION;

static	char	switch_help[] =
"Usage: irc [switches] [nickname] [server list] \n\
  The [nickname] can be at most 9 characters long\n\
  The [server list] is a whitespace separate list of server name\n\
  The [switches] may be any or all of the following\n\
   -c <channel>\tjoins <channel> o startup\n\
   -p <port>\tdefault server connection port (usually 6667)\n\
   -f\t\tyour terminal uses flow controls (^S/^Q), so IRCII shouldn't\n\
   -F\t\tyour terminal doesn't use flow control (default)\n\
   -s\t\tdon't use separate server processes (ircserv)\n\
   -S\t\tuse separate server processes (ircserv)\n\
   -d\t\truns IRCII in \"dumb\" terminal mode\n\
   -q\t\tdoes not load ~/.ircrc\n\
   -a\t\tadds default servers and command line servers to server list\n";

static	char	switch_help_l[] =
"   -l <file>\tloads <file> in place of your .ircrc\n\
   -L <file>\tloads <file> in place of your .ircrc and expands $ expandos\n";

/* Someone took this out... I stuck it back in -- Jake [WinterHawk] Khuon */
static char	switch_help_b[] =
	"   -b\t\tforce ircII to fork and return you to shell\n";

/* irc_exit: cleans up and leaves */
void irc_exit _((void))
{
	char *fooie = (char *) 0;
	do_hook(EXIT_LIST,"Exiting");
#ifdef WINTRHAWK
        /* Make the QUITcomment look a little less boring	-- Jake [WinterHawk] Khuon */
	malloc_sprintf(&fooie, "ircII %s taking fire!  Mayday Mayday Mayday!!! Eject Eject Eject!!!", IRCII_VERSION);
	close_server(-1, fooie);
#else
	malloc_sprintf(&fooie, "%s: Only the very best...", IRCII_VERSION);
	close_server(-1, fooie);
#endif
	new_free(&fooie);
	logger(0);
	set_history_file((char *) 0);
	clean_up_processes();
	if (!dumb)
	{
		cursor_to_input();	/* Needed so that ircII doesn't gobble
					 * the last line of the kill. */
		term_cr();
		if (term_clear_to_eol())
			term_space_erase(0);
		term_reset();
#if defined(_HPUX_SOURCE) || defined(ESIX)
		endwin();		/* Added for curses */
#ifdef ESIX
		system("tput reset");
		new_stty("sane");
#endif /* ESIX */
#endif /* defined(_HPUX_SOURCE) || defined(ESIX) */
	}
	exit(0);
}

RETSIGTYPE child_reap (int sig)
{
	check_wait_status(-1);
}

#ifdef CORECATCH
/* sigsegv: something to handle segfaults in a nice way */
#ifdef __STDC__
RETSIGTYPE coredump (int sig)
#else
RETSIGTYPE coredump(sig)
	int	sig;
#endif
{
	printf("IRCII has been terminated by a SIG%s\n\r", signals[sig]);
	printf("Please inform the EPIC project (ircii@rush.cc.edu) of this\n\r");
	printf("with as much detail as possible about what you were doing when it happened.\n\r");
	printf("Please include the version of IRCII (%s) and type of system in the report.\n\r", irc_version);
	printf("Please do not bother phone until we have decided it is not our error.\n\r");
	fflush(stdout);
	irc_exit();
}
#endif

/*
 * quit_response: Used by irc_io when called from irc_quit to see if we got
 * the right response to our question.  If the response was affirmative, the
 * user gets booted from irc.  Otherwise, life goes on. 
 */
#ifdef __STDC__
static void quit_response (char *dummy, char *ptr)
#else
static	void quit_response(dummy, ptr)
	char	*dummy;
	char	*ptr;
#endif
{
	int	len,
		old_irc_io_loop;

	old_irc_io_loop = irc_io_loop;
	irc_io_loop = 0;
	if ((len = strlen(ptr)) != 0)
	{
		if (!my_strnicmp(ptr, "yes", len))
		{
			send_to_server("QUIT :ircII %s -- Bigger and better then they said it could be done!", IRCII_VERSION);
			irc_exit();
		}
	}
	irc_io_loop = old_irc_io_loop;
}

/* irc_quit: prompts the user if they wish to exit, then does the right thing */
void
irc_quit _((void))
{
	static	int in_it = 0;

	if (in_it)
		return;
	in_it = 1;
	add_wait_prompt("Do you really want to quit? ", quit_response,
		empty_string, WAIT_PROMPT_LINE);
	in_it = 0;
}

/*
 * cntl_c: emergency exit.... if somehow everything else freezes up, hitting
 * ^C five times should kill the program. 
 */
RETSIGTYPE
cntl_c _((void))
{
#ifdef SYSVSIGNALS
	(void) MY_SIGNAL(SIGINT, cntl_c, 0);
#endif /* SYSVSIGNALS */
	if (cntl_c_hit++ >= 4)
		irc_exit();
}

RETSIGTYPE
sig_user1 _((void))
{
#ifdef SYSVSIGNALS
	(void) MY_SIGNAL(SIGUSR1, sig_user1, 0);
#endif /* SYSVSIGNALS */
	say("Got SIGUSR1, closing DCC connections and EXECed processes");
	close_all_dcc();
	close_all_exec();
}

#if DO_USER2
RETSIGTYPE
sig_user2 _((void))
{
#ifdef SYSVSIGNALS
	(void) MY_SIGNAL(SIGUSR2, sig_user2, 0);
#endif /* SYSVSIGNALS */
	say("Got SIGUSR2, jumping to normal loop");
	longjmp(outta_here);
}
#endif 

#ifdef MUNIX
/* A characteristic of PCS MUNIX - Ctrl-Y produces a SIGQUIT */
RETSIGTYPE
cntl_y _((void))
{
	(void) MY_SIGNAL(SIGQUIT, cntl_y, 0);
	edit_char((char) 25); /* Ctrl-Y */
}
#endif

static	void	show_version _((void))
{
#ifdef WINTRHAWK
	printf("ircII %s [Internal Release Code: %s]\n\r", irc_version, internal_version);
#else
	printf("ircII %s (Date of release: %s)\n\r",irc_version,internal_version);
#endif
	exit (0);
}

/*
 * parse_args: parse command line arguments for irc, and sets all initial
 * flags, etc. 
 *
 * major rewrite 12/22/94 -jfn
 *
 *
 * Im going to break backwards compatability here:  I think that im 
 * safer in doing this becuase there are a lot less shell script with
 * the command line flags then there are ircII scripts with old commands/
 * syntax that would be a nasty thing to break..
 *
 * Sanity check:
 *   Supported flags: -b, -l, -v, -c, -p, -f, -F, -L, -a, -S, -z
 *   New (changed) flags: -s, -I, -i, -n
 *
 * Rules:
 *   Each flag must be included by a hyphen:  -lb <filename> is not the
 * 		same as -l <filename> -b  any more...
 *   Each flag may or may not have a space between the flag and the argument.
 *   		-lfoo  is the same as -l foo
 *   Anything surrounded by quotation marks is honored as one word.
 *   The -c, -p, -L, -l, -s, -z flags all take arguments.  If no arguments
 *		are given between the flag and the next flag, an error
 * 		message is printed and the program is halted.
 *		Exception: the -s flag will be accepted without a argument.
 *		(ick: backwards compatability sucks. ;-)
 *   Arguments occuring after a flag that does not take an argument
 * 		will be parsed in the following way: the first instance
 *		will be an assumed nickname, and the second instance will
 *		will be an assumed server. (some semblance of back compat.)
 *   The -bl sequence will emit a depreciated feature warning.
 *   The -I flag forces you to become invisible <NOT YET SUPPORTED>
 *   The -i flag forces you to become visible <NOT YET SUPPORTED>
 *   The -s flag means "server" with an arg, "no ircserv" without
 *   The -n flag means "nickname"
 */
#ifdef __STDC__
static	char	*parse_args (char *argv[], int argc)
#else
static	char	*
parse_args(argv, argc)
	char	*argv[];
	int	argc;
#endif
{

	int ac;
	int add_servers = 0;
	struct passwd *entry;
	struct hostent *hp;

	char *channel = (char *) 0;
	char *ptr = (char *) 0;
	*nickname = 0;

	for ( ac = 1; ac < argc; ac++ )
	{
		if (argv[ac][0] == '-')
		{
		    switch (argv[ac][1]) {

			case 'v': /* Output ircII version */
			{
				show_version();
				/* NOTREACHED */
			}

			case 'c': /* Default channel to join */
			{
				char *what;

				if (argv[ac][2])
					what = &(argv[ac][2]);
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
				{
					fprintf(stderr, "Missing paramater after -c\n");
					exit(1);
				}
				malloc_strcpy(&channel, what);
				break;
			}

			case 'p': /* Default port to use */
			{
				char *what;
				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
				{
					fprintf(stderr, "Missing paramater after -p\n");
					exit(1);
				}
				irc_port = atoi(what);
				break;
			}

			case 'f': /* Use flow control */
			{
				use_flow_control = 1;
				if (argv[ac][2])
					fprintf(stderr, "Ignoring junk after -f\n");
				break;
			}

			case 'F': /* dont use flow control */
			{
				use_flow_control = 0;
				if (argv[ac][2])
					fprintf(stderr, "Ignoring junk after -F\n");
				break;
			}

			case 'd': /* use dumb mode */
			{
				dumb = 1;
				if (argv[ac][2])
					fprintf(stderr, "Ignoring junk after -d\n");
				break;
			}

			case 'l': /* Load some file instead of ~/.ircrc */
			{
				char *what;

				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
				{
					fprintf(stderr, "Missing argument to -l\n");
					exit(1);
				}
				malloc_strcpy(&ircrc_file, what);
				break;
			}

			case 'L': /* load and expand */
			{
				char *what;

				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
				{
					fprintf(stderr, "Missing argument to -L\n");
					exit(1);
				}
				malloc_strcpy(&ircrc_file, what);
				malloc_strcat(&ircrc_file, " -");
				break;
			}

			case 'a': /* add server, not replace */
			{
				add_servers = 1;
				if (argv[ac][2])
					fprintf(stderr, "Ignoring junk after -a\n");
				break;
			}

			case 'q': /* quick startup -- no .ircrc */
			{
				quick_startup = 1;
				if (argv[ac][2])
					fprintf(stderr, "Ignoring junk after -q\n");
				break;
			}

			case 's':
			{
				char *what = (char *) 0;
				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
					using_server_process = 0;
				if (what)
					build_server_list(what);
				break;
			}
			case 'S':
			{
				char *what = (char *) 0;
				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
					using_server_process = 1;
				if (what)
					build_server_list(what);
				break;
			}
			case 'b':
			{
/*	pull here -> */
				if (argv[ac][2] && argv[ac][2] != 'l')
					fprintf(stderr, "Ignoring junk after -b\n");
				else if (argv[ac][2] == 'l')
				{
					fprintf(stderr, "Usage of -bl is decprecated: use -b -l instead.\n");
					exit(1);
				}
				dumb = 1;
				use_input = 0;
				background = 1;
/* <-	tear to here */
				break;
			}

			case 'n':
			{
				char *what;
				if (argv[ac][2])
					what = &(argv[ac][2]);
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
				{
					fprintf(stderr,"Missing argument for -n\n");
					exit(1);
				}
				strcpy(nickname, what);
				break;
			}
			case 'z':
			{
				char *what;
				if (argv[ac][2])
					what = &argv[ac][2];
				else if (argv[ac+1] && argv[ac+1][0] != '-')
				{
					what = argv[ac+1];
					ac++;
				}
				else
					break;
				strmcpy(username, what, NAME_LEN);
				break;
			}

			case '\0': break;	/* ignore - alone */

			default:
			{
				fprintf(stderr, "Unknown flag: %s\n",argv[ac]);
				fprintf(stderr, "%s%s%s", switch_help, switch_help_b, switch_help_l);
				exit(1);
			}
		   } /* End of switch */
		}
		else
		{
			if (*nickname)
				build_server_list(argv[ac]);
			else
				strmcpy(nickname, argv[ac], NICKNAME_LEN);
		}
	}
	malloc_strcpy(&args_str, buffer);
	/* v-- right there was a '!' that should not have been there. */
	if ((ptr = getenv("IRCLIB")))
	{
		malloc_strcpy(&irc_lib, ptr);
		malloc_strcat(&irc_lib, "/");
	}
	else
		malloc_strcpy(&irc_lib, IRCLIB);

	if (!ircrc_file && (ptr = getenv("IRCRC")))
		malloc_strcpy(&ircrc_file, ptr);

	if (!*nickname && (ptr = getenv("IRCNICK")))
		strmcpy(nickname, ptr, NICKNAME_LEN);

	if ((ptr = getenv("IRCUMODE")))
		malloc_strcpy(&send_umode, ptr);

	if ((ptr = getenv("IRCNAME")))
		strmcpy(realname, ptr, REALNAME_LEN);

	if ((ptr = getenv("IRCPATH")))
		malloc_strcpy(&irc_path, ptr);
	else
	{
#ifdef IRCPATH
		malloc_strcpy(&irc_path, IRCPATH);
#else
		malloc_strcpy(&irc_path, ".:~/.irc:");
		malloc_strcat(&irc_path, irc_lib);
		malloc_strcat(&irc_path, "script");
#endif
	}

	set_string_var(LOAD_PATH_VAR, irc_path);
	new_free(&irc_path);
	if ((ptr = getenv("IRCSERVER")))
		build_server_list(ptr);
	if (!server_list_size() || add_servers)
	{
#ifdef SERVERS_FILE
		if (read_server_file() || (server_list_size() == 0))
#endif
		{
			char *ptr = (char *) 0;

			malloc_strcpy(&ptr, DEFAULT_SERVER);
			build_server_list(ptr);
			new_free(&ptr);
		}
	}
	if ((entry = getpwuid(getuid())))
	{
		if ((*realname == (char) 0) && entry->pw_gecos && *(entry->pw_gecos))
		{
#ifdef GECOS_DELIMITER
			if ((ptr = index(entry->pw_gecos, GECOS_DELIMITER)))
				*ptr = (char) 0;
#endif /* GECOS_DELIMITER */
			strmcpy(realname, entry->pw_gecos, REALNAME_LEN);
		}
		if (entry->pw_name && *(entry->pw_name) && (*username==(char) 0))
			strmcpy(username, entry->pw_name, NAME_LEN);
		if (entry->pw_dir && *(entry->pw_dir))
			malloc_strcpy(&my_path, entry->pw_dir);
	}
	if ((ptr = getenv("HOME")))
		malloc_strcpy(&my_path, ptr);
	else if (*my_path == (char) 0)
		malloc_strcpy(&my_path, "/");
	if (*realname == (char) 0)
		strmcpy(realname, "*Unknown*", REALNAME_LEN);

 #if !defined(I_AM_A_FASCIST_BASTARD) && defined(WINTRHAWK)
       /*
        * Yes... this is EXACTLY what you think it is.  And if you don't know..
        * then I'm not about to tell you!           -- Jake [WinterHawk] Khuon
        */
       else if (ptr = getenv("IRCUSER")) strmcpy(username, ptr, NAME_LEN);
       else if (ptr = getenv("USER")) strmcpy(username, ptr, NAME_LEN);
       else strmcpy(username, "Unknown", NAME_LEN);
#endif /* !defined(I_AM_A_FASCIST_BASTARD) && defined(WINTRHAWK) */
 
	gethostname(MyHostName, sizeof(MyHostName));
	if ((hp = gethostbyname(MyHostName)))
	{
		bcopy(hp->h_addr, (char *) &MyHostAddr, sizeof(MyHostAddr));
		local_ip_address.s_addr = ntohl(MyHostAddr.s_addr);
	}
	if (*nickname == (char) 0)
		strmcpy(nickname, username, NICKNAME_LEN);
	if (!check_nickname(nickname))
	{
		fprintf(stderr, "Illegal nickname %s\n", nickname);
		exit(1);
	}
	if (ircrc_file == (char *) 0)
	{
		ircrc_file = (char *) new_malloc(strlen(my_path) +
			strlen(IRCRC_NAME) + 1);
		strcpy(ircrc_file, my_path);
		strcat(ircrc_file, IRCRC_NAME);
	}
	return (channel);
}

/*
 * TimerTimeout:  Called from irc_io to help create the timeout
 * part of the call to select.
 */
time_t
TimerTimeout _((void))
{
	time_t	current;
	time_t	timeout_in;

	if (!PendingTimers)
		return 70; /* Just larger than the maximum of 60 */
	time(&current);
	timeout_in = PendingTimers->time - current;
	return (timeout_in < 0) ? 0 : timeout_in;
}

extern waiting;

/*
 * irc_io: the main irc input/output loop.   Handles all io from keyboard,
 * server, exec'd processes, etc.  If a prompt is specified, it is displayed
 * in the input line and cannot be backspaced over, etc. The func is a
 * function which will take the place of the SEND_LINE function (which is
 * what happens when you hit return at the end of a line). This function must
 * decide if it's ok to exit based on anything you really want.  It can then
 * set the global irc_io_loop to false to cause irc_io to exit. 
 *
 * This should be rewritten.
 */
#ifdef __STDC__
int irc_io (char *prompt, void (*func) (), int use_input, int loop)
#else
int irc_io(prompt, func, use_input, loop)
	char	*prompt;
	void	(*func) ();
	int	use_input;
	int	loop;
#endif
{
	static	int	level = 0;
	fd_set	rd,
		wd;
	char	buffer[BIG_BUFFER_SIZE + 1];	/* buffer much bigger than
						 * IRCD_BUFFER_SIZE */
	struct	timeval cursor_timeout,
		clock_timeout,
		right_away,
		timer,
		*timeptr;
	int	hold_over;
	int	old_loop;
	char	*last_input = NULL;
	char	*last_prompt = NULL;
	void	(*last_func) ();
	int	one_key = 0;
	Screen	*screen,
		*old_current_screen;

	last_func = get_send_line();
	if (use_input == -1)
		one_key = 1, prompt = NULL;
#ifdef	PRIV_PORT_ULC
	seteuid(getuid());
#endif
	/* time before cursor jumps from display area to input line */
	cursor_timeout.tv_usec = 0L;
	cursor_timeout.tv_sec = 1L;

	/* time delay for updating of internal clock */
	clock_timeout.tv_usec = 0L;
	clock_timeout.tv_sec = 15L;

	right_away.tv_usec = 0L;
	right_away.tv_sec = 0L;

	timer.tv_usec = 0L;

	old_loop = irc_io_loop;
	irc_io_loop = loop;

	if (level++ > 5)
	{
		level--;
		irc_io_loop = old_loop;
		return (1);
	}
	if (!dumb)
	{
		if (use_input)
		{
			malloc_strcpy(&last_input, get_input());
			set_input(empty_string);
			last_func = get_send_line();
			change_send_line(func);
		}
		if (prompt)
		{
			malloc_strcpy(&last_prompt, get_input_prompt());
			set_input_prompt(prompt);
		}
	}
	/*
	 * Here we work out if this has been called recursively or
	 * not..  and if not so.. -phone
	 */

#if defined(DEBUG) || defined(DO_USER2)
	if (level != 1)
	{
#ifdef DEBUG
		yell("--- Recursive call to irc_io() - careful");
#endif /* DEBUG */
	}
	else
	{
#if DO_USER2
		if (setjmp(outta_here))
			yell("*** Got SIGUSR2, Aborting");
#endif /* DO_USER2 */
	}
#endif /* DEBUG || DO_USER2 */

	timeptr = &clock_timeout;
	do
	{
		break_io_processing = 0;
		FD_ZERO(&rd);
		FD_ZERO(&wd);
		if (use_input)
			for (screen = screen_list;screen; screen = screen->next)
				if (screen->alive)
					FD_SET(screen->fdin, &rd);
		set_process_bits(&rd);
		set_server_bits(&rd);
		set_dcc_bits(&rd, &wd);
		if (term_reset_flag)
		{
			refresh_screen(NULL, NULL);
			term_reset_flag = 0;
		}
		timer.tv_sec = TimerTimeout();
		if (timer.tv_sec <= timeptr->tv_sec)
			timeptr = &timer;
		if ((hold_over = unhold_windows()) != 0)
			timeptr = &right_away;
		Debug(("irc_io: selecting with %l.%l timeout", timeptr->tv_sec,
			timeptr->tv_usec));
		switch (new_select(&rd, &wd, timeptr))
		{
		case 0:
		case -1:
       /*
        * yay for the QNX socket manager... drift here, drift there, oops,
        * i fell down a hole..
        */
#ifdef __QNX__
                        if (errno == EBADF || errno == ESRCH)
                                irc_io_loop = 0;
#endif
 
			if (cntl_c_hit)
			{
				if (one_key)
				{
					irc_io_loop = 0;
					break;
				}
				edit_char('\003');
				cntl_c_hit = 0;
			}
			if (!hold_over)
				cursor_to_input();
			break;
		default:
			if (term_reset_flag)
			{
				refresh_screen(NULL, NULL);
				term_reset_flag = 0;
			}
			old_current_screen = current_screen;
			set_current_screen(last_input_screen);
			if (!break_io_processing)
				dcc_check(&rd);
			if (!break_io_processing)
				do_server(&rd);
			set_current_screen(old_current_screen);
			for (screen = screen_list; screen &&
				!break_io_processing; screen = screen->next)
			{
				if (!screen->alive)
					continue;
				set_current_screen(screen);
				if (FD_ISSET(screen->fdin, &rd))
				{

	/*
	 * This section of code handles all in put from the terminal(s).
	 * connected to ircII.  Perhaps the idle time *shouldn't* be 
	 * reset unless its not a screen-fd that was closed..
	 *
	 * This section indented - phone, jan 1993
	 */

			idle_time = time(0);
			if (dumb)
			{
				int     old_timeout;

				old_timeout = dgets_timeout(1);
				if (dgets(buffer, INPUT_BUFFER_SIZE,
						screen->fdin, (char *) 0))
				{
					(void) dgets_timeout(old_timeout);
					if (one_key)
					{
						irc_io_loop = 0;
						break;
					}
					*(buffer + strlen(buffer) - 1) = '\0';
					if (get_int_var(INPUT_ALIASES_VAR))	
						parse_line(NULL, buffer,
						    empty_string, 1, 0);
					else
						parse_line(NULL, buffer,
						    NULL, 1, 0);
				}
				else
				{
					say("IRCII exiting on EOF from stdin");
					irc_exit();
				}
			}
			else
			{
				int server;
				char	loc_buffer[BIG_BUFFER_SIZE + 1];
				int	n, i;

				server = from_server;
				from_server = get_window_server(0);
				last_input_screen = screen;
				if (one_key)
				{
					if (read(screen->fdin, buffer, 1))
					{
						irc_io_loop = 0;
						break;
					}
#ifdef WINDOW_CREATE
					else
					{
						if (!is_main_screen(screen))
							kill_screen(screen);
						else
							irc_exit();
					}
#endif
				}
				else if ((n = read(screen->fdin, loc_buffer,
						BIG_BUFFER_SIZE)) != 0)
					for (i = 0; i < n; i++)
						edit_char(loc_buffer[i]);
		/*
		 * if the current screen isn't the main  screen,
		 * then the socket to the current screen must have
		 * closed, so we call kill_screen() to handle 
		 * this - phone, jan 1993.
		 * but not when we arent running windows - Fizzy, may 1993
		 * if it is the main screen we got an EOF on, we exit..
		 * closed tty -> chew cpu -> bad .. -phone, july 1993.
		 */
#ifdef WINDOW_CREATE
				else
				{
					if (!is_main_screen(screen))
						kill_screen(screen);
					else
						irc_exit();
				}
#endif /* WINDOW_CREATE */
				cntl_c_hit = 0;
				from_server = server;
			}

		/* End of intendation */

				}
			}
			set_current_screen(old_current_screen);
			if (!break_io_processing)
				do_processes(&rd);
			break;
		}
		ExecuteTimers();
		check_process_limits();
		check_wait_status(-1);
/*		if ((primary_server == -1) && !never_connected)
			do_hook(DISCONNECT_LIST, "%s", nickname); */
		timeptr = &clock_timeout;

		old_current_screen = current_screen;
		for (current_screen = screen_list; current_screen;
				current_screen = current_screen->next)
			if (current_screen->alive && is_cursor_in_display())
				timeptr = &cursor_timeout;
		set_current_screen(old_current_screen);

		if (update_clock(0))
		{
			if (get_int_var(CLOCK_VAR) || check_mail_status())
			{
				status_update(1);
				cursor_to_input();
			}
			if (primary_server != -1)
				do_notify();
		}
	}
	while (irc_io_loop);
	level--;
	irc_io_loop = old_loop;
	if (! dumb)
	{
		if (use_input)
		{
			set_input(last_input);
			new_free(&last_input);
			change_send_line(last_func);
		}
		if (prompt)
		{
			if (level == 0)
			    set_input_prompt(get_string_var(INPUT_PROMPT_VAR));
			else
			    set_input_prompt(last_prompt);
			new_free(&last_prompt);
		}
	}
	update_input(UPDATE_ALL);
	return (0);
}

static void check_password _((void))
{
#if defined(PASSWORD) && (defined(HARD_SECURE) || defined(SOFT_SECURE))
	char input_passwd[15];
#ifdef HAVE_GETPASS
	strcpy(input_passwd, getpass("Passwd: "));
#else
	fprintf(stderr, "Passwd: ");
	fgets(input_passwd, 14, stdin);
	input_passwd[strlen(input_passwd)-1] = 0;
#endif
	if (strcmp(input_passwd, PASSWORD))
		execl(SPOOF_PROGRAM, SPOOF_PROGRAM, NULL);
	else
		bzero(input_passwd, 14);
#endif
	return;
}

static void check_valid_user _((void))
{
#ifdef INVALID_UID_FILE
{
	long myuid = getuid();
	long curr_uid;
	char curr_uid_s[10];
	char *curr_uid_s_ptr;
	FILE *uid_file;

	uid_file = fopen(INVALID_UID_FILE, "r");
	if (uid_file == NULL)
		return;
	fgets(curr_uid_s, 9, uid_file);
	curr_uid_s[strlen(curr_uid_s)-1] = 0;
	while (!feof(uid_file))
	{
		curr_uid_s_ptr = curr_uid_s;
		while (!isdigit(*curr_uid_s_ptr) && *curr_uid_s_ptr != 0)
			curr_uid_s_ptr++;
		if (*curr_uid_s_ptr == 0)
		{
			fgets(curr_uid_s, 9, uid_file);
			continue;
		}
		else
			curr_uid = atol(curr_uid_s);
		if (myuid == curr_uid)
			exit(1);
		fgets(curr_uid_s, 9, uid_file);
		curr_uid_s[strlen(curr_uid_s)-1] = 0;
	}
	fclose(uid_file);
}
#endif
#ifdef HARD_SECURE
{
	int myuid = getuid();
	long curr_uid;
	char *curr_uid_s;
	char *uid_s_copy = NULL;
	char *uid_s_ptr;

	malloc_strcpy(&uid_s_copy, VALID_UIDS);
	uid_s_ptr = uid_s_copy;
	while (uid_s_ptr && *uid_s_ptr)
	{
		curr_uid_s = next_arg(uid_s_ptr, &uid_s_ptr);
		if (curr_uid_s && *curr_uid_s)
			curr_uid = atol(curr_uid_s);
		else
			continue;
		if (myuid == curr_uid)
			return;
	}
	new_free(&uid_s_ptr);
	execl(SPOOF_PROGRAM, SPOOF_PROGRAM, NULL);
}
#else
# if defined(SOFT_SECURE) && defined(VALID_UID_FILE)
{
	long myuid = getuid();
	long curr_uid;
	char curr_uid_s[10];
	char *uid_s_ptr;
	char *curr_uid_s_ptr;
	FILE *uid_file;

	uid_file = fopen(VALID_UID_FILE, "r");
	if (uid_file == NULL)
		return;
	fgets(curr_uid_s, 9, uid_file);
	curr_uid_s[strlen(curr_uid_s)-1] = 0;
	while (!feof(uid_file))
	{
		curr_uid_s_ptr = curr_uid_s;
		while (*curr_uid_s_ptr && !isdigit(*curr_uid_s_ptr))
			curr_uid_s_ptr++;
		if (*curr_uid_s_ptr == 0)
		{
			fgets(curr_uid_s, 9, uid_file);
			curr_uid_s[strlen(curr_uid_s)-1] = 0;
			continue;
		}
		else	
			curr_uid = atol(curr_uid_s);
		if (myuid == curr_uid)
		{
			fclose(uid_file);
			return;
		}
		fgets(curr_uid_s, 9, uid_file);
		curr_uid_s[strlen(curr_uid_s)-1] = 0;
	}
	fclose(uid_file);
	execl(SPOOF_PROGRAM, SPOOF_PROGRAM, NULL);
}
# endif
#endif
	return;
}


/*ARGSUSED*/
#ifdef __STDC__
void main (int argc, char *argv[], char *envp[])
#else
void main(argc, argv, envp)
	int	argc;
	char	*argv[];
	char	*envp[];
#endif
{
	char	*channel;

        start_time = time((time_t *)0);
	check_password();
	check_valid_user();
	channel = parse_args(argv, argc);
#if defined(_HPUX_SOURCE) || defined(ESIX)
	/* Curses code added for HP-UX use */
	if (!dumb)
	{
		initscr();
		noecho();
		cbreak();
	}
#endif /* _HPUX_SOURCE || ESIX */
	if ((use_input == 0) && !no_fork)
	{
		if (fork())
			_exit(0);
	}
#ifdef ESIX
	if (gethostname(hostname, NAME_LEN) == NULL)
#else
	if (gethostname(hostname, NAME_LEN))
#endif /* ESIX */
	{
		fprintf(stderr, "irc: couldn't figure out the name of your machine!\n");
		exit(1);
	}
	if (dumb)
		new_window();
	else
	{
		init_screen();
#if !defined(MUNIX) && !defined(_RT) && !defined(ESIX)
		(void) MY_SIGNAL(SIGCONT, term_cont, 0);
#endif /* !defined(MUNIX) && !defined(_RT) && !defined(ESIX) */
#if !defined(_RT) && defined(SIGWINCH)
		(void) MY_SIGNAL(SIGWINCH, refresh_screen, 0);
#endif /* _RT */
#ifndef ALLOC_DEBUG
# ifdef CORECATCH
		(void) MY_SIGNAL(SIGSEGV, coredump, 0);
#  ifdef SIGBUS
		(void) MY_SIGNAL(SIGBUS, coredump, 0);
#  endif
# else
		(void) MY_SIGNAL(SIGSEGV, SIG_DFL, 0);
		/* Linux doesn't have SIGBUS */
#  ifndef SIGBUS
		(void) MY_SIGNAL(SIGBUS, SIG_DFL, 0);
#  endif /* SIGBUS */
# endif /* CORECATCH */
#endif /* ALLOC_DEBUG */
#ifdef MUNIX
		(void) MY_SIGNAL(SIGQUIT, cntl_y, 0);
#endif
/* DONT CHANGE THIS OR HORDES OF LOCUSTS WILL EAT YOUR GRANDFATHER'S 
 * CORN CROP NEXT YEAR!!!!!!!!
 */
		if (background)
		(void) MY_SIGNAL(SIGHUP, SIG_IGN, 0);
		else
		(void) MY_SIGNAL(SIGHUP, irc_exit, 0);

		(void) MY_SIGNAL(SIGTERM, irc_exit, 0);
		(void) MY_SIGNAL(SIGPIPE, SIG_IGN, 0);
(void) MY_SIGNAL(SIGCHLD, child_reap, 0);
		(void) MY_SIGNAL(SIGINT, cntl_c, 0);
		(void) MY_SIGNAL(SIGALRM, SIG_IGN, 0);
#ifdef SIGSTOP
		(void) MY_SIGNAL(SIGSTOP, SIG_IGN, 0);
#endif
		(void) MY_SIGNAL(SIGUSR1, sig_user1, 0);
#if DO_USER2
		(void) MY_SIGNAL(SIGUSR2, sig_user2, 0);
#endif
	}
#ifdef _HPUX_SOURCE
	new_stty("opost");
#endif /* _HPUX_SOURCE */

	init_keys_1();
	init_variables();

	if (!dumb)
	{
		build_status((char *) 0);
		update_input(UPDATE_ALL);
	}

#ifdef MOTD_FILE
	{
		struct	stat	motd_stat,
				my_stat;
		char	*motd = NULL;
		int	des;

		malloc_strcpy(&motd, irc_lib);
		malloc_strcat(&motd, MOTD_FILE);
		if (stat_file(motd, &motd_stat) == 0)
		{
			strmcpy(buffer, my_path, BIG_BUFFER_SIZE);
			strmcat(buffer, "/.ircmotd", BIG_BUFFER_SIZE);
			if (stat_file(buffer, &my_stat))
			{
				my_stat.st_atime = 0L;
				my_stat.st_mtime = 0L;
			}
			unlink(buffer);
			if ((des = open(buffer, O_CREAT, S_IREAD | S_IWRITE)) != -1)
				close(des);
#ifndef ALWAYS_SHOW_MOTD
			if (motd_stat.st_mtime > my_stat.st_mtime)
			{
#endif
				put_file(motd);
		/* Thanks to Mark Dame <mdame@uceng.ec.edu> for this one */

#if PAUSE_AFTER_MOTD
		input_pause("********  Press any key to continue  ********");
#endif
				clear_window_by_refnum(0);
#ifndef ALWAYS_SHOW_MOTD
			}
#endif
		}
		new_free(&motd);
	}
#endif /* MOTD_FILE */

	global_all_off[0] = ALL_OFF;
	global_all_off[1] = '\0';
	get_connected(0);
	if (channel)
	{
		set_channel_by_refnum(0, channel);
		add_channel(channel, primary_server);
		new_free(&channel);
	}
	idle_time = time(0);
	set_input(empty_string);
        while (1)
	 irc_io(get_string_var(INPUT_PROMPT_VAR), NULL, use_input, irc_io_loop);
	irc_exit();
	return;
}
